import os
import pandas as pd
from tqdm.auto import tqdm
import torch
from pathlib import Path
import argparse
import fastremap
import time
import numpy as np



parser = argparse.ArgumentParser()
parser.add_argument("-d_p", "--data_path", type=str, default=r"../datasets")
parser.add_argument("-o_f", "--output_folder", type=str, default="Results")
parser.add_argument("-m_p", "--model_path", type=str, default=r"../models")
parser.add_argument("-m_f", "--model_folder", type=str)
parser.add_argument("-d", "--device", type=str, default=torch.device("cuda:0" if torch.cuda.is_available() else "cpu"))
parser.add_argument("-db", "--debug", default=False, type=lambda x: (str(x).lower() == 'true'))
parser.add_argument("-save_ims", "--save_ims", default=False, type=lambda x: (str(x).lower() == 'true'))
parser.add_argument("-data", "--dataset", type=str, default="segmentation", help="Name of the dataset to load")
parser.add_argument('-source', '--source_dataset', default=None, type=str)
parser.add_argument('-o_h', '--optimize_hyperparameters', default=False, type=lambda x: (str(x).lower() == 'true'),help="Optimize postprocessing parameters")
parser.add_argument('-tta', '--tta', default=False, type=lambda x: (str(x).lower() == 'true'),help="Test time augmentations")
parser.add_argument('-target', '--target_segmentation', default=None, type=str,help=" Cells or nuclei or both? Accepts: C,N, NC")
parser.add_argument('-params', '--params', default="default", type=str, help="Either 'default' or 'best_params'")
parser.add_argument('-window', '--window_size', default=128, type=int)
parser.add_argument('-set', '--test_set', default="Validation", type=str, help = "Validation or Test or Train")
parser.add_argument('-export_to_torchscript', '--export_to_torchscript', default=False,type=lambda x: (str(x).lower() == 'true'))
parser.add_argument('-export_to_bioimageio', '--export_to_bioimageio', default=False,type=lambda x: (str(x).lower() == 'true'))


#@timer
def instanseg_inference(val_images, val_labels, model, postprocessing_fn, device, parser_args, output_path, params=None,
                        instanseg=None, tta=False):
    
    if tta:
        import ttach as tta

        transforms = tta.Compose([
            tta.HorizontalFlip(),
            tta.VerticalFlip(),
            tta.Rotate90(angles=[0, 180, 90, 270]),
        ])

    from instanseg.utils.tiling import instanseg_padding, recover_padding
    count = 0
    time_dict = {'preprocessing': 0, 'model': 0, 'postprocessing': 0, 'torchscript': 0, 'combined': []}

    pred_masks = []
    gt_masks = []

    model.eval()

    #####
    #warmup
    #####
    imgs, _ = Augmenter.to_tensor(val_images[0], normalize=False)
    imgs = imgs.to(device)
    imgs, _ = Augmenter.normalize(imgs)
    with torch.no_grad():
        imgs, pad = instanseg_padding(imgs, extra_pad=0, min_dim=32)
        with torch.cuda.amp.autocast():
            pred = model(imgs[None,])
        pred = pred.float()
        pred = recover_padding(pred, pad).squeeze(0)
        if params is not None:
            with torch.cuda.amp.autocast():
                lab = postprocessing_fn(pred, **params, window_size=parser_args.window_size)
        else:
            with torch.cuda.amp.autocast():
                lab = postprocessing_fn(pred, img=imgs, window_size=parser_args.window_size)
        lab = lab.cpu().numpy()

        if tta:
            lab = method.TTA_postprocessing(imgs[None,], model, transforms, device=device)

    with torch.no_grad():
        for imgs, masks in tqdm(zip(val_images, val_labels), total=len(val_images)):

            torch.cuda.synchronize()
            start = time.time()
            imgs, masks = Augmenter.to_tensor(imgs, masks, normalize=False)
            imgs = imgs.to(device)
            imgs, _ = Augmenter.normalize(imgs)
            time_dict["preprocessing"] += time.time() - start
            torch.cuda.synchronize()
            start = time.time()

            if not tta:
                imgs, pad = instanseg_padding(imgs, extra_pad=0, min_dim=32, ensure_square=False)
                with torch.cuda.amp.autocast():
                    pred = model(imgs[None,])

                pred = recover_padding(pred, pad).squeeze(0)
                imgs = recover_padding(imgs, pad).squeeze(0)
                torch.cuda.synchronize()

                model_time = time.time() - start
                time_dict["model"] += model_time

                start = time.time()

                if params is not None:
                    with torch.cuda.amp.autocast():
                        lab = postprocessing_fn(pred, **params, window_size=parser_args.window_size)
                else:
                    with torch.cuda.amp.autocast():
                        lab = postprocessing_fn(pred, img=imgs, window_size=parser_args.window_size)

                torch.cuda.synchronize()

                postprocessing_time = time.time() - start

                time_dict["postprocessing"] += postprocessing_time

                time_dict["combined"].append({"time": model_time + postprocessing_time, "dimension": imgs.shape,
                                              "num_instances": len(torch.unique(lab) - 1)})

            else:
                if params is not None:
                    lab = method.TTA_postprocessing(imgs[None,], model, transforms, **params,
                                                    window_size=parser_args.window_size, device=device)
                else:
                    lab = method.TTA_postprocessing(imgs[None,], model, transforms, window_size=parser_args.window_size,
                                                    device=device)

            imgs = imgs.cpu().numpy()

            if isinstance(masks, torch.Tensor):
                masks = masks.numpy()
            if isinstance(lab, torch.Tensor):
                lab = lab.cpu().numpy()

            count += 1

            lab, _ = fastremap.renumber(lab, in_place=True)
            masks[masks > 0], _ = fastremap.renumber(masks[masks > 0], in_place=True)

            pred_masks.append(lab.astype(np.int16))
            gt_masks.append(masks.astype(np.int16))

            if parser_args.save_ims:
                from instanseg.utils.augmentations import Augmentations
                augmenter = Augmentations()
                display = augmenter.colourize(torch.tensor(imgs), random_seed=1)[0]

                def overlay(img, gt, color=None):
                    return save_image_with_label_overlay(img, gt, return_image=True, alpha=0.8,
                                                         label_boundary_mode="thick", label_colors=color)

                show_images(overlay(display.numpy(), torch.tensor(lab)),
                            save_str=output_path / str("images/" "overlay" + str(count)))

    print("Time spent in preprocessing", time_dict["preprocessing"], "Time spent in model:", time_dict["model"],
          "Time spent in postprocessing:", time_dict["postprocessing"])

    return pred_masks, gt_masks, time_dict


if __name__ == "__main__":

    from instanseg.utils.utils import show_images, save_image_with_label_overlay, _move_channel_axis
    from instanseg.utils.model_loader import load_model
    from instanseg.utils.metrics import compute_and_export_metrics
    from instanseg.utils.augmentations import Augmentations

    parser_args = parser.parse_args()
    if parser_args.model_folder == "None":
        parser_args.model_folder = ""

    model, model_dict = load_model(path=parser_args.model_path, folder=parser_args.model_folder)

    data_path = Path(parser_args.data_path)
    os.environ["INSTANSEG_DATASET_PATH"] = str(parser_args.data_path)
    device = parser_args.device
    n_sigma = model_dict['n_sigma']

    model_path = Path(parser_args.model_path) / Path(parser_args.model_folder)

    parser_args.loss_function = model_dict['loss_function']
    if parser_args.source_dataset is None:
        parser_args.source_dataset = model_dict['source_dataset']

    if parser_args.loss_function.lower() == "instanseg_loss":
        from instanseg.utils.loss.instanseg_loss import InstanSeg

        method = InstanSeg(binary_loss_fn_str=model_dict["binary_loss_fn"], seed_loss_fn=model_dict["seed_loss_fn"],
                           n_sigma=model_dict["n_sigma"],
                           cells_and_nuclei=model_dict["cells_and_nuclei"], to_centre=model_dict["to_centre"],
                           window_size=parser_args.window_size, dim_coords=model_dict["dim_coords"],
                           feature_engineering_function=model_dict["feature_engineering"])

        if parser_args.target_segmentation is None:
            parser_args.cells_and_nuclei = model_dict["cells_and_nuclei"]
            parser_args.target_segmentation = model_dict["target_segmentation"]

        else:
            if len(parser_args.target_segmentation) == 2:
                parser_args.cells_and_nuclei = True
            else:
                parser_args.cells_and_nuclei = False

        parser_args.pixel_size = model_dict["pixel_size"]

        import math

        if math.isnan(parser_args.pixel_size):
            parser_args.pixel_size = None

        method.initialize_pixel_classifier(model)


        def loss_fn(*args, **kwargs):
            return method.forward(*args, **kwargs)


        def get_labels(pred, **kwargs):
            return method.postprocessing(pred, **kwargs, device=device, max_seeds=10000)


        dim_out = method.dim_out

    else:
        raise NotImplementedError("Loss function not recognized", parser_args.loss_function)

    model.eval()
    if "inference_folder" not in parser_args or parser_args.inference_folder is None:
        from instanseg.utils.data_loader import _read_images_from_pth

        val_images, val_labels, val_meta = _read_images_from_pth(args=parser_args, sets=[parser_args.test_set],
                                                                 dataset=parser_args.dataset)

    else:
        from instanseg.utils.data_loader import _read_images_from_path

        val_images, val_labels = _read_images_from_path(sets=[parser_args.test_set])

    datasets_str = np.unique([item['parent_dataset'] for item in val_meta])
    print("Datasets used:", datasets_str)

    model.to(device)

    output_path = model_path / parser_args.output_folder
    if not os.path.exists(output_path):
        os.mkdir(output_path)
    if parser_args.save_ims:
        if not os.path.exists(output_path / "images"):
            os.mkdir(output_path / "images")

    torch.cuda.empty_cache()
    count = 0


    #  pdb.set_trace()

    Augmenter = Augmentations(dim_in=model_dict['dim_in'], shape=None,
                              channel_invariant=model_dict['channel_invariant'])

    val_data = [Augmenter.duplicate_grayscale_channels(*Augmenter.to_tensor(img, label, normalize= True)) for img, label in
            zip(val_images, val_labels)]

   # val_data = [(img, label) for img, label in zip(val_images, val_labels)]
    # val_data = [Augmenter.normalize(img,label,percentile=0.) for img, label in val_data]

    if parser_args.pixel_size is not None and parser_args.pixel_size != "None":
        print("Warning, rescaling image and ground truth labels to pixel size:", parser_args.pixel_size,
              "microns/pixel")

        for i, (img, label) in enumerate(val_data):
            if "pixel_size" not in val_meta[i].keys():
                val_meta[i]["pixel_size"] = 0.5

            elif val_meta[i]["pixel_size"] == "pixel_size":
                val_meta[i]["pixel_size"] = 0.5  #bug in mesmer dataset

        val_data = [Augmenter.torch_rescale(img, label, current_pixel_size=val_meta[i]['pixel_size'],
                                            requested_pixel_size=parser_args.pixel_size,
                                            modality=val_meta[i]["image_modality"], crop=False) for i, (img, label) in
                    enumerate(val_data)]

    #val_data = [Augmenter.colourize(img,label,c_nuclei = val_meta[i]['nuclei_channels'][0]) for i, (img,label) in enumerate(val_data)]

    # from instanseg.utils.augmentations import get_marker_location
    # val_meta = [get_marker_location(meta) for meta in val_meta]
    # val_data = [Augmenter.extract_nucleus_and_cytoplasm_channels(img,label,c_nuclei = val_meta[i]['nuclei_channels'][0],metadata = val_meta[i]) for i, (img,label) in enumerate(val_data)]

    val_images = [item[0] for item in val_data]  #[::-1]
    val_labels = [item[1] for item in val_data]  #[::-1]

    from instanseg.utils.utils import count_instances

    freq = np.array([count_instances(label) for label in val_labels])
    area = np.array(
        [(len(label[label > 0].flatten())) / f for f, label in
         zip(freq, val_labels)])  # this will break for images with 0 labels

    print("Found:", sum(freq), "instances, across", len(freq), "images.", "Median area:", np.median(area), "pixels")

    if parser_args.optimize_hyperparameters:
        from instanseg.utils.AI_utils import optimize_hyperparameters

        params = optimize_hyperparameters(model, postprocessing_fn=method.postprocessing, val_images=val_images,
                                          val_labels=val_labels, verbose=True)
        pd.DataFrame.from_dict(params, orient='index').to_csv(output_path / "best_params.csv",
                                                              header=False)
    else:
        if parser_args.params == "default":
            params = None
        else:
            df = pd.read_csv(output_path / "best_params.csv", header=None)
            params = {row[0]: row[1] for row in df.values}

    # params["window_size"] = parser_args.window_size
    instanseg = None
    if parser_args.export_to_torchscript:
        from instanseg.utils.utils import export_to_torchscript

        print("Exporting model to torchscript")
        export_to_torchscript(parser_args.model_folder)
        instanseg = torch.jit.load("../torchscripts/" + parser_args.model_folder + ".pt")
    if parser_args.export_to_bioimageio:
        print("Exporting model to bioimageio")
        from instanseg.utils.create_bioimageio_model import export_bioimageio

        instanseg = torch.jit.load("../torchscripts/" + parser_args.model_folder + ".pt")
        export_bioimageio(instanseg, deepimagej=True, test_img_path="../examples/HE_example.tif",
                          model_name=parser_args.model_folder)

    pred_masks, gt_masks, time_dict = instanseg_inference(val_images,
                                                          val_labels,
                                                          model,
                                                          postprocessing_fn=get_labels,
                                                          device=device,
                                                          parser_args=parser_args,
                                                          output_path=output_path,
                                                          params=params,
                                                          instanseg=instanseg,
                                                          tta=parser_args.tta)

    pd.DataFrame(time_dict['combined']).to_csv(output_path / "timing_dict.csv", header=True)

    if parser_args.cells_and_nuclei:
        pred_nuclei_masks = [pred_mask[0] for gt_mask, pred_mask in zip(gt_masks, pred_masks) if gt_mask[0].min() >= 0]
        gt_nuclei_masks = [gt_mask[0] for gt_mask, pred_mask in zip(gt_masks, pred_masks) if gt_mask[0].min() >= 0]

        pred_cell_masks = [pred_mask[1] for gt_mask, pred_mask in zip(gt_masks, pred_masks) if gt_mask[1].min() >= 0]
        gt_cell_masks = [gt_mask[1] for gt_mask, pred_mask in zip(gt_masks, pred_masks) if gt_mask[1].min() >= 0]

        compute_and_export_metrics(gt_nuclei_masks, pred_nuclei_masks, output_path, target="Nuclei")
        compute_and_export_metrics(gt_cell_masks, pred_cell_masks, output_path, target="Cells")
    else:
        pred_masks = [(pred_mask).squeeze()[None] for gt_mask, pred_mask in zip(gt_masks, pred_masks) if
                      gt_mask.min() >= 0]
        gt_masks = [(gt_mask).squeeze()[None] for gt_mask, pred_mask in zip(gt_masks, pred_masks) if gt_mask.min() >= 0]
        compute_and_export_metrics(gt_masks, pred_masks, output_path,
                                   target="Cells" if parser_args.target_segmentation == "C" else "Nuclei")




import os
import numpy as np
import matplotlib.pyplot as plt
from tqdm.auto import tqdm
import sys

from torch import nn
import torch
import torch.optim as optim



import argparse
from pathlib import Path
import pandas as pd
import pdb

#torch.autograd.set_detect_anomaly(True) #For debugging cuda errors

parser = argparse.ArgumentParser()

#basic usage
parser.add_argument("-d_p", "--data_path", type=str, default=r"../datasets", help="Path to the .pth file")
parser.add_argument("-data", "--dataset", type=str, default="segmentation", help="Name of the dataset to load")
parser.add_argument('-source', '--source_dataset', default="all", type=str, help = "Which datasets to use for training. Input is 'all' or a list of datasets (e.g. [TNBC_2018,LyNSeC,IHC_TMA,CoNSeP])")
parser.add_argument("-m_f", "--model_folder", type=str, default=None, help = "Name of the model to resume training. This must be a folder inside model_path")
parser.add_argument("-m_p", "--model_path", type=str, default=r"../models", help = "Path to the folder containing the models")
parser.add_argument("-o_p", "--output_path", type=str, default=r"../models", help = "Path to the folder where the results will be saved")
parser.add_argument("-e_s", "--experiment_str", type=str, default="my_first_instanSeg", help = "String to identify the experiment")
parser.add_argument("-d", "--device", type=str, default=torch.device("cuda:0" if torch.cuda.is_available() else "cpu"))
parser.add_argument('-num_workers', '--num_workers', default=3, type=int, help = "Number of CPU cores to use for data loading")
parser.add_argument('-ci', '--channel_invariant', default=False, type=lambda x: (str(x).lower() == 'true'), help = "Whether to add a channel invariant model to the pipeline")
parser.add_argument('-target', '--target_segmentation', default="N",type=str, help = " Cells or nuclei or both? Accepts: C,N, NC")  
parser.add_argument('-pixel_size', '--requested_pixel_size', default=None, type=float, help = "Requested pixel size to rescale the input images")


#advanced usage
parser.add_argument("-bs", "--batch_size", type=int, default=3)
parser.add_argument("-e", "--num_epochs", type=int, default=500)
parser.add_argument('-len_epoch', '--length_of_epoch', default=1000, type=int, help = "Number of samples per epoch")
parser.add_argument("-lr", "--lr", type=float, default=0.001, help = "Learning rate")
parser.add_argument("-m", "--model_str", type=str, default="InstanSeg_UNet", help = "Model backbone to use")
parser.add_argument("-s", "--save", type=bool, default=True, help = "Whether to save model outputs every time a new best F1 score is achieved")
parser.add_argument("-l_fn", "--loss_function", type=str, default='instanseg_loss', help = "Method to use for segmentation, only instanseg_loss is supported")
parser.add_argument("-n_sigma", "--n_sigma", type=int, default=2, help = "Number of sigma channels, must be at least 1")
parser.add_argument("-cluster", "--on_cluster", type=bool, default=False, help ="Flag to disable tqdm progress bars and other non-essential outputs, useful for running on a cluster")
parser.add_argument("-w", "--weight", default=False, type = lambda x: (str(x).lower() == 'true'), help = "Weight the random sampler in the training set to oversample images with more instances")
parser.add_argument("-layers", "--layers", type=str, default="[32, 64, 128, 256]", help = "UNet layers")
parser.add_argument("-slice", "--data_slice", type=int, default=None, help = "Slice of the dataset to use, useful for debugging (e.g. only train on 1 image)")
parser.add_argument("-clip", "--clip", type=float, default=20, help ="Gradient clipping value")
parser.add_argument("-decay", "--weight_decay", type=float, default=0.000, help = "Weight decay")
parser.add_argument("-drop", "--dropprob", type=float, default=0., help = "Dropout probability, Not implemented for InstanSeg_UNet")
parser.add_argument("-tf", "--transform_intensity", type=float, default=0.5, help = "Intensity transformation factor")
parser.add_argument("-dim_in", "--dim_in", type=int, default=3,help="Number of channels that the (backbone) model expects. This is also the number of channels a channel invariant model would output.")
parser.add_argument("-dummy", "--dummy", default=False, type=lambda x: (str(x).lower() == 'true'),help="Use the training set as a validation set, this will trigger a warning message. use only for debugging")
parser.add_argument('-to_centre', '--to_centre', default=False, type=lambda x: (str(x).lower() == 'true'), help = "Whether to use the instance centroid or the learnt instance centroid in InstanSeg")
parser.add_argument('-multi_centre', '--multi_centre', default=True, type=lambda x: (str(x).lower() == 'true'), help = "Allow multi centres per instance, uses local maxima algorithm")
parser.add_argument('-open_license', '--open_license', default=False, type=lambda x: (str(x).lower() == 'true'), help = "Whether to filter out images that do not have an open license during training")
parser.add_argument('-modality', '--image_modality', default="all", type=str, help = "Filter out images that do not have this modality: Brightfield, Fluorescence, all")
parser.add_argument('-binary_loss_fn', '--binary_loss_fn', default="lovasz_hinge", type=str, help = "Loss function to use for instance segmentation: lovasz_hinge or dice_loss are supported. lovasz_hinge is a lot slower to start converging")
parser.add_argument('-seed_loss_fn', '--seed_loss_fn', default="l1_distance", type=str, help = "Loss function to use for seed selection, only binary_xloss and l1_distance are supported. Binary_xloss is much faster, but l1_distance is usually more accurate")
parser.add_argument('-anneal', '--cosineannealing', default=False, type=lambda x: (str(x).lower() == 'true'), help = "Whether to use cosine annealing for the learning rate")
parser.add_argument('-o_h', '--optimize_hyperparameters', default=False, type=lambda x: (str(x).lower() == 'true'), help = "Whether to optimize hyperparameters every 10 epochs")
parser.add_argument('-hotstart', '--hotstart_training', default=10, type=int, help = "Number of epochs to train the model with binary_xloss before starting the main training loop (default=10)")
parser.add_argument('-window', '--window_size', default=128, type=int, help = "Size of the window containing each instance")
parser.add_argument('-multihead', '--multihead', default= True, type=lambda x: (str(x).lower() == 'true'), help = "Whether to branch the decoder into multiple heads.")
parser.add_argument('-dim_coords', '--dim_coords', default=2, type=int, help = "Dimensionality of the coordinate system. Little support for anything but 2")
parser.add_argument('-norm', '--norm', default="BATCH", type=str, help = "Norm layer to use: None, INSTANCE, INSTANCE_INVARIANT, BATCH")
parser.add_argument('-mlp_w', '--mlp_width', default=5, type=int, help = "Width of the MLP hidden dim")
parser.add_argument('-feature_engineering', '--feature_engineering', default="0", type=str, help = "Feature engineering function to use")
parser.add_argument('-augmentation_type', '--augmentation_type', default="minimal", type=str, help = "'minimal' or 'heavy' or 'brightfield_only'")
parser.add_argument('-adaptor_net', '--adaptor_net_str', default="1", type=str, help = "Adaptor net to use")

parser.add_argument('-rng_seed', '--rng_seed', default=None, type=int, help = "Optional seed for the random number generator")
parser.add_argument('-use_deterministic', '--use_deterministic', default=False, type=lambda x: (str(x).lower() == 'true'), help = "Whether to use deterministic algorithms (default=False)")


def main(model, loss_fn, train_loader, test_loader, num_epochs=1000, epoch_name='output_epoch'):
    global best_f1_score

    train_losses = []
    test_losses = []

    best_f1_score = -1
    f1_list = []
    f1_list_cells = []

    for epoch in range(num_epochs):

        print("Epoch:", epoch)

        train_loss, train_time = train_epoch(model, device, train_loader, loss_fn, optimizer, args = args, writer = writer)

        if epoch <= 5 + 500 * (
                1 - args.on_cluster) and not args.model_folder:  # Training is just starting AND we are not loading a model
            save_epoch_outputs = True
        else:
            save_epoch_outputs = False

        test_loss, f1_score, test_time = test_epoch(model, device, test_loader, loss_fn, debug=False,
                                                    best_f1=best_f1_score,
                                                    save_bool=save_epoch_outputs,
                                                    args = args,
                                                    postprocessing_fn = method.postprocessing,
                                                    method = method,
                                                    iou_threshold = iou_threshold,
                                                    writer = writer,
                                                    save_str=str(
                                                        args.output_path / str(
                                                            f"epoch_outputs/{epoch_name}_" + str(epoch))))
        train_losses.append(train_loss)
        test_losses.append(test_loss)

        if epoch % 10 ==0 and args.optimize_hyperparameters:
            best_params = optimize_hyperparameters(model,postprocessing_fn = method.postprocessing,data_loader= test_loader,verbose = not args.on_cluster, show_progressbar = not args.on_cluster)
            method.update_hyperparameters(best_params)


        dict_to_print = {"train_loss": train_loss, "test_loss": test_loss, "training_time": int(train_time),
                         "testing_time": int(test_time)}

        if args.cells_and_nuclei:
            f1_list.append(f1_score[0])
            f1_list_cells.append(f1_score[1])
            dict_to_print["f1_score_nuclei"] = f1_score[0]
            dict_to_print["f1_score_cells"] = f1_score[1]
            f1_score = np.nanmean(f1_score)
            dict_to_print["f1_score_joint"] = f1_score

        else:
            f1_score = f1_score[0]
            f1_list.append(f1_score)
            dict_to_print["f1_score"] = f1_score

        if args.cosineannealing:
            dict_to_print["lr:"] = optimizer.param_groups[0]["lr"]
            scheduler.step()

                
        if f1_score > best_f1_score or save_epoch_outputs:
            best_f1_score = np.maximum(f1_score, best_f1_score)

            print("Saving model, best f1_score:", best_f1_score)

            torch.save({
                'f1_score': best_f1_score,
                'epoch': epoch,
                'model_state_dict': model.state_dict(),
                'optimizer_state_dict': optimizer.state_dict(),
            }, args.output_path / "model_weights.pth") 


        print(str(dict_to_print).replace("{", "").replace("}", "").replace("'", ""))

    return model, train_losses, test_losses, f1_list, f1_list_cells


if __name__ == "__main__":

    from instanseg.utils.utils import plot_average, _choose_device
    from instanseg.utils.AI_utils import optimize_hyperparameters, train_epoch, test_epoch
    from instanseg.utils.model_loader import build_model_from_dict, load_model_weights
    from instanseg.utils.data_loader import _read_images_from_pth, get_loaders

    args = parser.parse_args()

    
    args.data_path = Path(args.data_path)

    if not os.path.exists(args.output_path):
        os.mkdir(args.output_path)

    args.output_path = Path(args.output_path) / args.experiment_str
    print("Saving results to {}".format(os.path.abspath(args.output_path)))
    if not os.path.exists(args.output_path):
        os.mkdir(args.output_path)

    os.environ["INSTANSEG_DATASET_PATH"] = str(args.data_path)
    os.environ["INSTANSEG_OUTPUT_PATH"] = str(args.output_path)

    # Seed as many rngs as we can
    if args.rng_seed:
        print(f'Setting RNG seed to {args.rng_seed}')
        torch.manual_seed(args.rng_seed)
        np.random.seed(args.rng_seed)
        import random
        random.seed(args.rng_seed)
    else:
        print('RNG seed not set')

    if args.use_deterministic:
        print('Setting use_deterministic_algorithms=True')
        torch.use_deterministic_algorithms(True)


    args.layers = eval(args.layers)
    args_dict = vars(args)
    num_epochs = args.num_epochs
    n_sigma = args.n_sigma
    on_cluster = args.on_cluster
    dim_in = args.dim_in

    if args.norm == "None":
        args.norm = None

    if len(args.target_segmentation) == 2:
        args.cells_and_nuclei = True
    else:
        args.cells_and_nuclei = False
        
    device = _choose_device(args.device)

    if args.loss_function == "instanseg_loss":
        from instanseg.utils.loss.instanseg_loss import InstanSeg


        method = InstanSeg(binary_loss_fn_str=args.binary_loss_fn, 
                        seed_loss_fn = args.seed_loss_fn, 
                        device = device,
                        n_sigma=n_sigma,
                        cells_and_nuclei=args.cells_and_nuclei, 
                        to_centre = args.to_centre, 
                        window_size = args.window_size, 
                        dim_coords= args.dim_coords, 
                        multi_centre= args.multi_centre, 
                        feature_engineering_function=args.feature_engineering )  # binary_xloss, lovasz_hinge dice_loss general_dice_loss

        def loss_fn(*args, **kwargs):
            return method.forward(*args, **kwargs)
        
        dim_out = method.dim_out

    else:
        raise NotImplementedError("Loss function not recognized", args.loss_function)

    args.dim_out = dim_out
    args_dict = vars(args)

    if int(dim_in) == 0:
        args_dict["dim_in"] = None
    else:
        args_dict["dim_in"] = int(dim_in)
    args_dict["dropprob"] = float(args.dropprob)

    model = build_model_from_dict(args_dict)
    # from fvcore.nn import FlopCountAnalysis
    # flops = FlopCountAnalysis(model, torch.randn(1,3,256,256))
    # print("Number of flops:",flops.total()/1e9)
    # from fvcore.nn import flop_count_str
    # print(flop_count_str(flops))

    if args.loss_function in ["instanseg_loss"]:
        from instanseg.utils.loss.instanseg_loss import has_pixel_classifier_model

        if not has_pixel_classifier_model(model):
            model = method.initialize_pixel_classifier(model, MLP_width = args.mlp_width)

    if args.model_folder:
        if args.model_folder == "None":
            args.model_folder = ""

        model, model_dict = load_model_weights(model, path=args.model_path, folder=args.model_folder, device=device, dict = args_dict)

        if not args.channel_invariant:
            optimizer = optim.Adam(model.parameters(), lr=args.lr, weight_decay=args.weight_decay)
            optimizer.load_state_dict(model_dict['optimizer_state_dict'])
            optimizer.param_groups[0]['lr'] = args.lr  # you may want to remove this line in the future

        print("Resuming training from epoch", model_dict['epoch'])

    else:
        optimizer = optim.Adam(model.parameters(), lr=args.lr, weight_decay=args.weight_decay)

    if args.channel_invariant:
        from instanseg.utils.models.ChannelInvariantNet import AdaptorNetWrapper, has_AdaptorNet
        if not has_AdaptorNet(model):
            model = AdaptorNetWrapper(model,adaptor_net_str=args.adaptor_net_str, norm = args.norm)
        optimizer = optim.Adam(model.parameters(), lr=args.lr, weight_decay=args.weight_decay)

    if args.cosineannealing:
        scheduler = torch.optim.lr_scheduler.CosineAnnealingLR(optimizer, T_max=100, eta_min=0.00001)


    if "[" in args.source_dataset:
        args.source_dataset = args.source_dataset.replace("[", "").replace("]", "").replace("'", "").split(",")
    else:
        args.source_dataset = args.source_dataset


    train_images, train_labels, train_meta, val_images, val_labels, val_meta = _read_images_from_pth(data_path = args.data_path, dataset = args.dataset, data_slice = args.data_slice, dummy = args.dummy, args = args, sets= ["Train","Validation"] )

    train_loader, test_loader = get_loaders(train_images, train_labels, val_images, val_labels, train_meta, val_meta, args)
    
    if torch.cuda.device_count() > 1:
        print("Using", torch.cuda.device_count(), "GPUs!")
        model = nn.DataParallel(model)

    model.to(device)

    if args.save:
        if not os.path.exists(args.output_path):
            os.mkdir(args.output_path)
        if not os.path.exists(args.output_path / "epoch_outputs"):
            os.mkdir(args.output_path / "epoch_outputs")


    pd.DataFrame.from_dict(args_dict, orient='index').to_csv(args.output_path / "experiment_log.csv",
                                                            header=False)
    
    iou_threshold = np.linspace(0.5, 1.0, 10)
    from torch.utils.tensorboard import SummaryWriter

    if not os.path.exists(Path("../tensorboard_logs/")):
        os.mkdir(Path("../tensorboard_logs/"))

    writer = SummaryWriter( Path("../tensorboard_logs/"+str(args.experiment_str)+'/gradient_monitoring'),flush_secs=10)


    if args.hotstart_training > 0:
        hot_epochs = args.hotstart_training
        print("Hotstart for "+str(hot_epochs)+" epochs with binary_xloss and dice_loss")
        method.update_seed_loss("binary_xloss")
        method.update_binary_loss("dice_loss")
        model, train_losses, test_losses, f1_list, f1_list_cells = main(model, loss_fn, train_loader, test_loader, num_epochs=hot_epochs, epoch_name='hotstart_epoch')

        print("Starting main training loop with",args.seed_loss_fn, "and", args.binary_loss_fn)
        method.update_seed_loss(args.seed_loss_fn)
        method.update_binary_loss(args.binary_loss_fn)



    model, train_losses, test_losses, f1_list, f1_list_cells = main(model, loss_fn, train_loader, test_loader, num_epochs=num_epochs)



    from instanseg.utils.model_loader import load_model
    model, model_dict = load_model(folder="", path=args.output_path) #Load model from checkpoint
    model.eval()
    model.to(device)

    df = pd.DataFrame({"train_loss": train_losses, "test_loss": test_losses, "f1_score": f1_list})
    df.to_csv(args.output_path / "experiment_metrics.csv", index=False, header=True)


    fig = plot_average(train_losses, test_losses, window_size=len(train_losses) // 10 + 1)
    plt.savefig(args.output_path / "loss.png")
    plt.close()

    if args.cells_and_nuclei:
        fig = plt.plot(f1_list, label="f1 score nuclei")
        plt.plot(f1_list_cells, label="f1 score cells")
        plt.ylim(0, 1)
        plt.savefig(args.output_path / "f1_metric.png")
        plt.legend()
        plt.close()

    else:
        fig = plt.plot(f1_list, label="f1 score")
        plt.ylim(0, 1)
        plt.savefig(args.output_path / "f1_metric.png")
        plt.close()

    if not args.on_cluster and args.experiment_str is None:
        experiment_str = "experiment"
    elif args.experiment_str is not None:
        experiment_str = args.experiment_str


import numpy as np
import torch
import pdb

from stardist import matching
import fastremap

from typing import Union
def _robust_f1_mean_calculator(nan_list: Union[list, np.ndarray]):
    nan_list = np.array(nan_list)
    if len(nan_list) == 0:
        return np.nan
    elif np.isnan(nan_list).all():
        return np.nan
    else:
        return np.nanmean(nan_list)


def _robust_average_precision(labels, predicted, threshold):

    for i in range(len(labels)):
        if labels[i].min() < 0 and not (labels[i] < 0).all():
            labels[i][labels[i] < 0] = 0 #sparse labels
            predicted[i][labels[i] < 0] = 0 

    if labels[0].shape[0] != 2: #cells or nuclei
        labels = [labels[i].detach().cpu().numpy().astype(np.int32) for i, l in enumerate(labels) if labels[i].min() >= 0 and labels[i].max() > 0]
        predicted = [predicted[i].detach().cpu().numpy().astype(np.int32) for i, l in enumerate(labels) if labels[i].min() >= 0 and labels[i].max() > 0]

        if len(labels)==0:
            return np.nan
        

        stats = matching.matching_dataset([l for l in labels], [p for p in predicted], thresh=threshold, show_progress = False)
        f1i = [stat.f1 for stat in stats]
        return _robust_f1_mean_calculator(f1i)
    else:
        f1is = [] 
        for i, _ in enumerate(["nuclei", "cells"]):
            labels_tmp = [fastremap.renumber(labels[j][i].detach().cpu().numpy())[0].astype(np.int32) for j, l in enumerate(labels) if labels[j][i].min() >= 0 and labels[j][i].max() > 0]
            predicted_tmp = [fastremap.renumber(predicted[j][i].detach().cpu().numpy())[0].astype(np.int32) for j, l in enumerate(labels) if labels[j][i].min() >= 0 and labels[j][i].max() > 0]

            if len(labels_tmp)==0:
                f1is.append(np.nan)
                continue

            stats = matching.matching_dataset([l for l in labels_tmp], [p for p in predicted_tmp],thresh=threshold, show_progress = False)
            f1i = [stat.f1 for stat in stats]


            f1is.append(_robust_f1_mean_calculator(f1i))

        return f1is
    



import pandas as pd
from tqdm import tqdm
def compute_and_export_metrics(gt_masks, pred_masks, output_path, target, return_metrics = False, show_progress = False, verbose = True):
    taus = [ 0.5, 0.6, 0.7, 0.8, 0.9]
    stats = [matching.matching_dataset(gt_masks, pred_masks, thresh=t, show_progress=False, by_image = False) for t in tqdm(taus, disable=not show_progress)]
    df_list = []

    for stat in stats:
        df_list.append(pd.DataFrame([stat]))
    df = pd.concat(df_list, ignore_index=True)

    mean_f1 = df[["thresh", "f1"]].iloc[:].mean()["f1"]
    mean_panoptic_quality = df[["thresh", "panoptic_quality"]].iloc[:].mean()["panoptic_quality"]
    panoptic_quality_05 = df[["thresh", "panoptic_quality"]].iloc[0]["panoptic_quality"]
    f1_05 = df[["thresh", "f1"]].iloc[0]["f1"]


    df["mean_f1"] = mean_f1
    df["f1_05"] = f1_05
    df["mean_PQ"] = mean_panoptic_quality
    df["SQ"] = panoptic_quality_05 / f1_05

    if verbose:
        print("Target:",target)
        print("Mean f1 score: ", mean_f1)
        print("f1 score at 0.5: ", f1_05)
        print("SQ: ", panoptic_quality_05 / f1_05)

    if return_metrics:
        return mean_f1, f1_05, panoptic_quality_05 / f1_05

    if output_path is not None:

        df.to_csv(output_path / str(target + "_matching_metrics.csv"))



from typing import Union, List, Optional
import numpy as np
import torch
from torch import nn
from torch.nn.functional import interpolate
from pathlib import Path, PosixPath

import sys
import pdb


def _to_tensor_float32(image: Union[np.ndarray, torch.Tensor]) -> torch.Tensor:
    if isinstance(image, np.ndarray):      
        if image.dtype == np.uint16:
            image = image.astype(np.int32)
        image = torch.from_numpy(image).float()
    
    image = image.squeeze()

    assert image.dim() <= 3 and image.dim() >= 2, f"Input image shape {image.shape()} is not supported."

    image = torch.atleast_3d(image)
    channel_index = np.argmin(image.shape) #Note, this could break for small, highly multiplexed images.
    if channel_index != 0:
        image = image.movedim(channel_index, 0)

    return image

def _rescale_to_pixel_size(image: torch.Tensor, 
                           requested_pixel_size: float, 
                           model_pixel_size: float) -> torch.Tensor:

    if image.squeeze().dim() == 3:
        not_batched = True
        image = image[None]
    else:
        not_batched = False

    scale_factor = requested_pixel_size / model_pixel_size

    if not np.allclose(scale_factor,1, 0.01):
        image = interpolate(image, scale_factor=scale_factor, mode="bilinear")
    
    if not_batched:
        image = image[0]

    return image

def display_colourized(mIF):
    from instanseg.utils.utils import _move_channel_axis, generate_colors

    mIF = _to_tensor_float32(mIF)
    mIF = mIF / (mIF.max() + 1e-6)
    if mIF.shape[0]!=3:
        colours = generate_colors(num_colors=mIF.shape[0])
        colour_render = (mIF.flatten(1).T @ torch.tensor(colours)).reshape(mIF.shape[1],mIF.shape[2],3)
    else:
        colour_render = mIF
    colour_render = torch.clamp_(colour_render, 0, 1)
    colour_render = _move_channel_axis(colour_render,to_back = True).detach().numpy()*255
    return colour_render.astype(np.uint8)


class InstanSeg():
    """
    Main class for running InstanSeg.
    """
    def __init__(self, 
                 model_type: Union[str,nn.Module] = "brightfield_nuclei", 
                 device: Optional[str] = None, 
                 image_reader: str = "tiffslide",
                verbosity = 1, #0,1,2
                ):
        
        """
        Parameters:
            model_type (Union[str,nn.Module])
                The type of model to use. If a string is provided, the model will be downloaded. 
                If the model is not public, it will look for a model in your bioimageio folder. 
                If an nn.Module is provided, this model will be used.

            device (Optional[str]): 
                The device to run the model on. If None, the device will be chosen automatically.

            image_reader (str): 
                The image reader to use. Options are "tiffslide", "skimage.io", "bioio", "AICSImageIO".

            verbosity (int): 
                The verbosity level. 0 is silent, 1 is normal, 2 is verbose.
        
        """
        from instanseg.utils.utils import download_model, _choose_device

        self.verbosity = verbosity
        self.verbose = verbosity != 0

        if isinstance(model_type, nn.Module):
            self.instanseg = model_type
        else:
            self.instanseg = download_model(model_type, verbose = self.verbose)
        self.inference_device = _choose_device(device, verbose= self.verbose)
        self.instanseg = self.instanseg.to(self.inference_device)

        self.prefered_image_reader = image_reader
        self.small_image_threshold = 3 * 1500 * 1500 #max number of image pixels to be processed on GPU.
        self.medium_image_threshold = 10000 * 10000 #max number of image pixels that could be loaded in RAM.
        self.prediction_tag = "_instanseg_prediction"

    def read_image(self, image_str: str):
        if self.prefered_image_reader == "tiffslide":
            from tiffslide import TiffSlide
            slide = TiffSlide(image_str)
            img_pixel_size = slide.properties['tiffslide.mpp-x']
            width,height = slide.dimensions[0], slide.dimensions[1]
            num_pixels = width * height
            if num_pixels < self.medium_image_threshold:
                image_array = slide.read_region((0, 0), 0, (width, height), as_array=True)
            else:
                return image_str, img_pixel_size
            
        elif self.prefered_image_reader == "skimage.io":
            from skimage.io import imread
            image_array = imread(image_str)
            img_pixel_size = None

        elif self.prefered_image_reader == "bioio":
            from bioio import BioImage
            slide = BioImage(image_str)
            img_pixel_size = slide.physical_pixel_sizes.X
            num_pixels = np.cumprod(slide.shape)[-1]
            if num_pixels < self.medium_image_threshold:
                image_array = slide.get_image_data().squeeze()
            else:
                return image_str, img_pixel_size

        elif self.prefered_image_reader == "AICSImageIO":
            from aicsimageio import AICSImage
            slide = AICSImage(image_str)
            img_pixel_size = slide.physical_pixel_sizes.X
            num_pixels = np.cumprod(slide.shape)[-1]
            if num_pixels < self.medium_image_threshold:
                image_array = slide.get_image_data().squeeze()
            else:
                return image_str, img_pixel_size
        else:
            raise NotImplementedError(f"Image reader {self.prefered_image_reader} is not implemented.")
        
        if img_pixel_size is None or float(img_pixel_size) < 0 or float(img_pixel_size) > 2:
            img_pixel_size = self.read_pixel_size(image_str)

        if img_pixel_size is not None:
            import warnings
            if float(img_pixel_size) <= 0 or float(img_pixel_size) > 2:
                warnings.warn(f"Pixel size {img_pixel_size} microns per pixel is invalid.")
                img_pixel_size = None

        return image_array, img_pixel_size
    
    def read_pixel_size(self,image_str: str):
        try:
            from tiffslide import TiffSlide
            slide = TiffSlide(image_str)
            img_pixel_size = slide.properties['tiffslide.mpp-x']
            if img_pixel_size is not None and img_pixel_size > 0 and img_pixel_size < 2:
                return img_pixel_size
        except Exception as e:
            print(e)
            pass
        from aicsimageio import AICSImage 
        try:
            
            slide = AICSImage(image_str)
            img_pixel_size = slide.physical_pixel_sizes.X
            if img_pixel_size is not None and img_pixel_size > 0 and img_pixel_size < 2:
                return img_pixel_size
        except Exception as e:
            print(e)
            pass
        from bioio import BioImage
        try:
            slide = BioImage(image_str)
            img_pixel_size = slide.physical_pixel_sizes.X
            if img_pixel_size is not None and img_pixel_size > 0 and img_pixel_size < 2:
                return img_pixel_size
        except Exception as e:
            print(e)
            pass
        import slideio
        try:
            slide = slideio.open_slide(image_str, driver = "AUTO")
            scene  = slide.get_scene(0)
            img_pixel_size = scene.resolution[0] * 10**6

            if img_pixel_size is not None and img_pixel_size > 0 and img_pixel_size < 2:
                    
                return img_pixel_size
        except Exception as e:
            print(e)
            pass
        print("Could not read pixel size from image metadata.")
        
        return None

    
    def read_slide(self, image_str: str):      
        if self.prefered_image_reader == "tiffslide":
            from tiffslide import TiffSlide
            slide = TiffSlide(image_str)
        # elif self.prefered_image_reader == "AICSImageIO":
        #     from aicsimageio import AICSImage
        #     slide = AICSImage(image_str)
        # elif self.prefered_image_reader == "bioio":
        #     from bioio import BioImage
        #     slide = BioImage(image_str)
        # elif self.prefered_image_reader == "slideio":
        #     import slideio
        #     slide = slideio.open_slide(image_str, driver = "AUTO")

        else:
            raise NotImplementedError(f"Image reader {self.prefered_image_reader} is not implemented for whole slide images.")
        return slide
    
    def _to_tensor(self, image: Union[np.ndarray, torch.Tensor]) -> torch.Tensor:
        return _to_tensor_float32(image)
    
    def _normalise(self, image: torch.Tensor) -> torch.Tensor:
        from instanseg.utils.utils import percentile_normalize, _move_channel_axis
        assert image.ndim == 3 or image.ndim == 4, f"Input image shape {image.shape} is not supported."
        if image.dim() == 3:
            image = percentile_normalize(image)
            image = image[None]
        else:
            image = torch.stack([percentile_normalize(i) for i in image])

        return image

    def eval(self, image: Union[str, List[str]], 
            pixel_size: Optional[float] = None,
            save_output: bool = False,
            save_overlay: bool = False,
            save_geojson: bool = False,
            **kwargs):
        """
        Evaluate the input image or list of images using the InstanSeg model.
        """

        if isinstance(image, PosixPath):
            image = str(image)
        if isinstance(image, str):
            initial_type = "not_list"
            image_list = [image]
        else:
            initial_type = "list"
            image_list = image
        
        output_list = []
    
        for image in image_list:
            image_array, img_pixel_size = self.read_image(image)

            if img_pixel_size is None and pixel_size is not None:
                img_pixel_size = pixel_size
            if img_pixel_size is None:
                import warnings
                warnings.warn("Pixel size not provided and could not be read from image metadata, this may lead to innacurate results.")
                
            if not isinstance(image_array, str):
                
                num_pixels = np.cumprod(image_array.shape)[-1]
                if num_pixels < self.small_image_threshold:
                    instances = self.eval_small_image(image = image_array, 
                                                       pixel_size = img_pixel_size, 
                                                       return_image_tensor=False, **kwargs)

                else:
                    instances = self.eval_medium_image(image = image_array, 
                                                       pixel_size = img_pixel_size, 
                                                       return_image_tensor=False, **kwargs)

                output_list.append(instances)

                if save_output:
                    self.save_output(image, instances, image_array = image_array, save_overlay = save_overlay, save_geojson = save_geojson)
     
                    
            else:
                self.eval_whole_slide_image(image_array, pixel_size, **kwargs)

        if initial_type == "not_list":
            output = output_list[0]
        else:
            output = output_list

        
        return output
    
    def save_output(self,image_path: str, 
                    labels: torch.Tensor,
                    image_array: Optional[np.ndarray] = None,
                    save_overlay = False,
                    save_geojson = False,):

        if isinstance(image_path, str):
            image_path = Path(image_path)
        if isinstance(labels, torch.Tensor):
            labels = labels.cpu().detach().numpy()

        new_stem = image_path.stem + self.prediction_tag

        from skimage import io

        if self.verbose:

            out_path = Path(image_path).parent / (new_stem + ".tiff")
            print(f"Saving output to {out_path}")
            io.imsave(out_path, labels.squeeze().astype(np.int32), check_contrast=False)

        if save_geojson:
            if labels.ndim == 3:
                labels = labels[None]

            output_dimension = labels.shape[1]
            from instanseg.utils.utils import labels_to_features
            import json
            if output_dimension == 1:
                features = labels_to_features(labels[0,0],object_type = "detection")

            elif output_dimension == 2:
                features = labels_to_features(labels[0,0],object_type = "detection",classification="Nuclei") + labels_to_features(labels[0,1],object_type = "detection",classification = "Cells")
            geojson = json.dumps(features)

            geojson_path = Path(image_path).parent / (new_stem + ".geojson")
            with open(os.path.join(geojson_path), "w") as outfile:
                outfile.write(geojson)
        
        if save_overlay:

            if self.verbose:
                out_path = Path(image_path).parent / (new_stem + "_overlay.tiff")
                print(f"Saving overlay to {out_path}")
            assert image_array is not None, "Image array must be provided to save overlay."
            display = self.display(image_array, labels)
            
            io.imsave(out_path, display, check_contrast=False)


            

    def eval_small_image(self,image: torch.Tensor, pixel_size: Optional[float] = None,
                          normalise: bool = True,
                          return_image_tensor: bool = True,
                          target = "all_outputs", #or "nuclei" or "cells"
                          rescale_output: bool = True,
                          **kwargs):
        """
        Evaluate the input image using the InstanSeg model.
        
        Args:
            image (torch.Tensor): The input image(s) to be evaluated.
            pixel_size (Optional[float]): The pixel size of the image. If not provided, it will be read from the image metadata.
        """
        from instanseg.utils.utils import percentile_normalize, _move_channel_axis

        image = _to_tensor_float32(image)

        original_shape = image.shape

        if pixel_size is not None:
            image = _rescale_to_pixel_size(image, pixel_size, self.instanseg.pixel_size)

            if original_shape[-2] != image.shape[-2] or original_shape[-1] != image.shape[-1]:
                img_has_been_rescaled = True
            else:
                img_has_been_rescaled = False

        image = image.to(self.inference_device)

        assert image.dim() ==3 or image.dim() == 4, f"Input image shape {image.shape} is not supported."

        if normalise:
            if image.dim() == 3:
                image = percentile_normalize(image)
                image = image[None]
            else:
                image = torch.stack([percentile_normalize(i) for i in image])

        
        if target != "all_outputs" and self.instanseg.cells_and_nuclei:
            assert target in ["nuclei", "cells"], "Target must be 'nuclei', 'cells' or 'all_outputs'."
            if target == "nuclei":
                target_segmentation = torch.tensor([1,0])
            else:
                target_segmentation = torch.tensor([0,1])
        else:
            target_segmentation = torch.tensor([1,1])

        with torch.amp.autocast('cuda'):

        
            instanseg_kwargs = {k: v for k, v in kwargs.items() if k not in ["batch_size", "tile_size", "normalisation_subsampling_factor"]}
            instances = self.instanseg(image,target_segmentation = target_segmentation, **instanseg_kwargs)

        if pixel_size is not None and img_has_been_rescaled and rescale_output:  
            instances = interpolate(instances, size=original_shape[-2:], mode="nearest")

            if return_image_tensor:
                image = interpolate(image, size=original_shape[-2:], mode="bilinear")
        if return_image_tensor:
            return instances.cpu(), image.cpu()
        else:
            return instances.cpu()

    def eval_medium_image(self,image: torch.Tensor, 
                          pixel_size: Optional[float] = None, 
                          normalise: bool = True,
                          tile_size: int = 512,
                          batch_size: int = 1,
                        return_image_tensor: bool = True,
                        normalisation_subsampling_factor: int = 1,
                        target = "all_outputs", #or "nuclei" or "cells"
                        rescale_output: bool = True,
                          **kwargs):
        
        from instanseg.utils.utils import percentile_normalize, _move_channel_axis

        image = _to_tensor_float32(image)
        
        from instanseg.utils.tiling import sliding_window_inference
        original_shape = image.shape

        if pixel_size is None:
            import warnings
            warnings.warn("Pixel size not provided, this may lead to innacurate results.")
        else:
            image = _rescale_to_pixel_size(image, pixel_size, self.instanseg.pixel_size)

            if original_shape[-2] != image.shape[-2] or original_shape[-1] != image.shape[-1]:
                img_has_been_rescaled = True
            else:
                img_has_been_rescaled = False
        assert image.dim() ==3, f"Input image shape {image.shape} is not supported."
        
        if normalise:
            image = percentile_normalize(image, subsampling_factor=normalisation_subsampling_factor)
            image = image

        output_dimension = 2 if self.instanseg.cells_and_nuclei else 1

        if target != "all_outputs" and output_dimension == 2:
            assert target in ["nuclei", "cells"], "Target must be 'nuclei', 'cells' or 'all_outputs'."
            if target == "nuclei":
                target_segmentation = torch.tensor([1,0])
            else:
                target_segmentation = torch.tensor([0,1])
            output_dimension = 1
        else:
            target_segmentation = torch.tensor([1,1])

        instances = sliding_window_inference(image,
                    self.instanseg, 
                    window_size = (tile_size,tile_size),
                    sw_device = self.inference_device,
                    device = 'cpu', 
                    batch_size= batch_size,
                    output_channels = output_dimension,
                    show_progress= self.verbose,
                    target_segmentation = target_segmentation,
                    **kwargs).float()
        
        if pixel_size is not None and img_has_been_rescaled and rescale_output:  
            instances = interpolate(instances, size=original_shape[-2:], mode="nearest")

            if return_image_tensor:
                image = interpolate(image, size=original_shape[-2:], mode="bilinear")

        if return_image_tensor:
            return instances.cpu(), image.cpu()
        else:
            return instances.cpu()
        
    def eval_whole_slide_image(self, image: str, pixel_size: Optional[float] = None, 
                               normalise: bool = True, 
                               batch_size: int = 1,
                               tile_size: int = 512,
                               output_geojson: bool = False,
                               **kwargs):
            from instanseg.utils.tiling import segment_image_larger_than_memory
                                 
            segment_image_larger_than_memory(instanseg_class= self,
                                            image_path= image, 
                                            memory_block_size = (int(self.medium_image_threshold**0.5),int(self.medium_image_threshold **0.5)),  #this is the size of the image that will be read in memory
                                            inference_tile_size = (tile_size,tile_size), #this is the size of the image that will be passed to the model
                                            batch_size= batch_size,
                                            to_geojson= output_geojson, 
                                            pixel_size =pixel_size,
                                            prediction_tag = self.prediction_tag, 
                                            sw_device = self.inference_device,
                                            normalisation_subsling_factor=10,
                                            **kwargs)
            
    
    def display(self, image: torch.tensor,
               instances: torch.Tensor):
        
        from instanseg.utils.utils import save_image_with_label_overlay

        if isinstance(image, torch.Tensor):
            image = image.cpu().detach().numpy()

        im_for_display = display_colourized(image.squeeze())

        output_dimension = instances.shape[1]

        if output_dimension ==1: #Nucleus or cell mask
            labels_for_display = instances[0,0] #Shape is 1,H,W
            image_overlay = save_image_with_label_overlay(im_for_display,lab=labels_for_display,return_image=True, label_boundary_mode="thick", label_colors=None,thickness=10,alpha=0.9)
        elif output_dimension ==2: #Nucleus and cell mask
            nuclei_labels_for_display = instances[0,0]
            cell_labels_for_display = instances[0,1] #Shape is 1,H,W
            image_overlay = save_image_with_label_overlay(im_for_display,lab=nuclei_labels_for_display,return_image=True, label_boundary_mode="thick", label_colors="red",thickness=10)
            image_overlay = save_image_with_label_overlay(image_overlay,lab=cell_labels_for_display,return_image=True, label_boundary_mode="inner", label_colors="green",thickness=1)

        return image_overlay
    

    def cluster_instances_by_mean_channel_intensity(self, image_tensor: torch.Tensor, labeled_output: torch.Tensor):

        #This is experimental code that is not yet implemented. You'll need to install rapids_singlecell, cuml and scanpy to run this code.

        from instanseg.utils.biological_utils import get_mean_object_features
        import fastremap
        import numpy as np
        from instanseg.utils.utils import apply_cmap
        from instanseg.utils.pytorch_utils import torch_fastremap
        import rapids_singlecell as rsc
        import scanpy as sc
        import matplotlib.pyplot as plt

        X_features = get_mean_object_features( image_tensor.to("cuda"), labeled_output.to("cuda"),)

        adata = sc.AnnData(X_features.cpu().numpy())
        rsc.get.anndata_to_GPU(adata)
        rsc.pp.scale(adata)
        rsc.pp.neighbors(adata, n_neighbors=8, n_pcs=50)
        rsc.tl.umap(adata)
        rsc.tl.leiden(adata, resolution=0.1)

        # Create the UMAP plot
        fig, axes = plt.subplots(1, 2, figsize=(15, 7))
        mapping = fastremap.component_map(np.arange(1, len(adata.obs["leiden"]) + 1), adata.obs["leiden"].astype(np.int64) + 1)
        labs = torch_fastremap(labeled_output[0, 0])
        labels = fastremap.remap(labs.numpy(), mapping, preserve_missing_labels=True)

        labels_disp = apply_cmap(labels, labels > 0, cmap="tab10")

        # Show the labeled image
        axes[0].imshow(labels_disp)
        axes[0].set_title('Labeled Image')
        axes[0].axis('off')

        sc.pl.umap(adata, color="leiden", legend_loc='on data', cmap="tab10", title='UMAP with Leiden Clustering', s=30, ax=axes[1], show = False)
        axes[1].axis('off')
        plt.subplots_adjust(wspace=0., hspace=0)
        plt.show()


            
if __name__ == "__main__":


    import os
    import pdb
    sys.path = sys.path[1:]
    from instanseg.utils.utils import show_images

    example_image_folder = Path(os.path.join(os.path.dirname(__file__),"./examples/"))

    instanseg_brightfield = InstanSeg("brightfield_nuclei")


    image_array, pixel_size = instanseg_brightfield.read_image(example_image_folder/"HE_example.tif")

    labeled_output, image_tensor  = instanseg_brightfield.eval_small_image(image_array, pixel_size)

    display = instanseg_brightfield.display(image_tensor, labeled_output)
    show_images(image_tensor,display, colorbar=False)


    instanseg_fluorescence = InstanSeg("fluorescence_nuclei_and_cells")
    image_array, pixel_size = instanseg_fluorescence.read_image("../instanseg/examples/LuCa1.tif")

    labeled_output, image_tensor  = instanseg_fluorescence.eval_small_image(image_array, pixel_size)

    display = instanseg_fluorescence.display(image_tensor, labeled_output)

    show_images(display, colorbar=False)

   # instanseg = InstanSeg("fluorescence_nuclei_and_cells")
   # image_array,pixel_size = instanseg.read_image(example_image_folder / "adam.ome.tif")

   # labeled_output, image_tensor  = instanseg.eval_medium_image(image_array, pixel_size, normalisation_subsampling_factor=10, batch_size=3)

   # instanseg.save_output(example_image_folder / "adam.ome.tif", labeled_output, image_array, output_overlay=True, output_geojson=True)

    # instanseg = InstanSeg("brightfield_nuclei")
    # # instances = instanseg.eval_whole_slide_image(example_image_folder / "HE_Hamamatsu.tiff")

    # instanseg = InstanSeg("brightfield_nuclei")
    # instances = instanseg.eval(example_image_folder / "HE_example.tif")

    # instanseg.prefered_image_reader = "skimage.io"
    # instances = instanseg.eval(example_image_folder / "HE_example.tif")

    # instanseg.prefered_image_reader = "AICSImageIO"
    # instances = instanseg.eval(example_image_folder / "HE_example.tif")

    # instanseg = InstanSeg("fluorescence_nuclei_and_cells")
    # instances = instanseg.eval(example_image_folder / "LuCa1.tif", batch_size=3)

    # show_images(instances)

        



        














 




import os
import torch
import torch.nn.functional as F

from typing import Tuple, List, Union




def remap_values(remapping: torch.Tensor, x: torch.Tensor) -> torch.Tensor:
    """
    This remaps the values in x according to the pairs in the remapping tensor.
    remapping: 2,N      Make sure the remapping is 1 to 1, and there are no loops (i.e. 1->2, 2->3, 3->1). Loops can be removed using graph based connected components algorithms (see instanseg postprocessing for an example)
    x: any shape
    """
    sorted_remapping = remapping[:, remapping[0].argsort()]
    index = torch.bucketize(x.ravel(), sorted_remapping[0])
    return sorted_remapping[1][index].reshape(x.shape)


# def torch_fastremap(x: torch.Tensor) -> torch.Tensor:
#    # if x.max() == 0:
#    #     return x
#   #  fg = x[x > 0]
#     unique_values = torch.unique(fg, sorted=True)
#     new_values = torch.arange(len(unique_values), dtype=x.dtype, device=x.device)
#     remapping = torch.stack((unique_values, new_values))
#     fg = remap_values(remapping, fg)
#     x[x > 0] = fg + 1
#     return x

def torch_fastremap(x: torch.Tensor) -> torch.Tensor:
    if x.max() == 0:
        return x
    unique_values = torch.unique(x, sorted=True)
    new_values = torch.arange(len(unique_values), dtype=x.dtype, device=x.device)
    remapping = torch.stack((unique_values, new_values))
    return remap_values(remapping, x)



def torch_onehot(x: torch.Tensor) -> torch.Tensor:
    # x is a labeled image of shape _,_,H,W returns a onehot encoding of shape 1,C,H,W

    if x.max() == 0:
        return torch.zeros_like(x).reshape(1, 0, *x.shape[-2:])
    H, W = x.shape[-2:]
    x = x.view(-1, 1, H, W)
    x = x.squeeze().view(1, 1, H, W)
    unique = torch.unique(x[x > 0])
    x = x.repeat(1, len(unique), 1, 1)
    return x == unique.unsqueeze(-1).unsqueeze(-1)


def fast_iou(onehot: torch.Tensor, threshold: float = 0.5) -> torch.Tensor:
    # onehot is C,H,W
    if onehot.ndim == 3:
        onehot = onehot.flatten(1)
    onehot = (onehot > threshold).float()
    intersection = onehot @ onehot.T
    union = onehot.sum(1)[None].T + onehot.sum(1)[None] - intersection
    return intersection / union


def fast_sparse_iou(sparse_onehot: torch.Tensor) -> torch.Tensor:

    intersection = torch.sparse.mm(sparse_onehot, sparse_onehot.T).to_dense()
    sparse_sum = torch.sparse.sum(sparse_onehot, dim=(1,))[None].to_dense()
    union = sparse_sum.T + sparse_sum - intersection
    return intersection / union



def instance_wise_edt(x: torch.Tensor, edt_type: str = 'auto') -> torch.Tensor:
    """
    Create instance-normalized distance map from a labeled image.
    Each pixel within an instance gives the distance to the closed background pixel,
    divided by the maximum distance (so that the maximum within an instance is 1).

    The calculation of the Euclidean Distance Transform can use the 'edt' or 'monai'
    packages.
    'edt' is faster for CPU computation, while 'monai' can use cucim for GPU acceleration
    where CUDA is available.
    Use 'auto' to decide automatically.
    """
    if x.max() == 0:
        return torch.zeros_like(x).squeeze()
    is_mps = x.is_mps
    if is_mps:
        # Need to convert to CPU for MPS, because distance transform gives float64 result
        # and Monai's internal attempt to convert type will fail
        x = x.to('cpu')

    use_edt = edt_type == 'edt' or (edt_type != 'monai' and not x.is_cuda)
    if use_edt:
        import edt
        xedt = torch.from_numpy(edt.edt(x[0].cpu().numpy(), black_border=False))
        x = torch_onehot(x)[0] * xedt.to(x.device)
    else:
        import monai
        x = torch_onehot(x)
        x = monai.transforms.utils.distance_transform_edt(x[0])

    # Normalize instance distances to have max 1
    x = x / (x.flatten(1).max(1)[0]).view(-1, 1, 1)
    x = x.sum(0)

    if is_mps:
        x = x.type(torch.FloatTensor).to('mps')
    return x



def fast_dual_iou(onehot1: torch.Tensor, onehot2: torch.Tensor) -> torch.Tensor:
    """
    Returns the intersection over union between two dense onehot encoded tensors
    """
    # onehot1 and onehot2 are C1,H,W and C2,H,W

    C1 = onehot1.shape[0]
    C2 = onehot2.shape[0]

    max_C = max(C1, C2)

    onehot1 = torch.cat((onehot1, torch.zeros((max_C - C1, *onehot1.shape[1:]))), dim=0)
    onehot2 = torch.cat((onehot2, torch.zeros((max_C - C2, *onehot2.shape[1:]))), dim=0)

    onehot1 = onehot1.flatten(1)
    onehot1 = (onehot1 > 0.5).float()  # onehot should be binary

    onehot2 = onehot2.flatten(1)
    onehot2 = (onehot2 > 0.5).float()

    intersection = onehot1 @ onehot2.T
    union = (onehot1).sum(1)[None].T + (onehot2).sum(1)[None] - intersection

    return (intersection / union)[:C1, :C2]


def torch_sparse_onehot(x: torch.Tensor, flatten: bool = False) -> Tuple[torch.Tensor, torch.Tensor]:
    # x is a labeled image of shape _,_,H,W returns a sparse tensor of shape C,H,W
    unique_values = torch.unique(x, sorted=True)
    x = torch_fastremap(x)

    H, W = x.shape[-2], x.shape[-1]


    if flatten:

        if x.max() == 0:
            return torch.zeros_like(x).reshape(1, 1, H*W)[:,:0] , unique_values
        

        x = x.reshape(H * W)
        xxyy = torch.nonzero(x > 0).squeeze(1)
        zz = x[xxyy] - 1
        C = x.max().int().item()

       # print(C, H, W, type(C), type(H), type(W))
        sparse_onehot = torch.sparse_coo_tensor(torch.stack((zz, xxyy)).long(), (torch.ones_like(xxyy).float()),
                                                size=(int(C), int(H * W)), dtype=torch.float32)

    else:
        if x.max() == 0:
            return torch.zeros_like(x).reshape(1, 0, H,W) , unique_values

        x = x.squeeze().view(H, W)
        x_temp= torch.nonzero(x > 0).T
        zz = x[x_temp[0], x_temp[1]] - 1
        C = x.max().int().item()
        sparse_onehot = torch.sparse_coo_tensor(torch.stack((zz, x_temp[0], x_temp[1])).long(), (torch.ones_like(x_temp[0]).float()),
                                                size=(int(C), int(H), int(W)), dtype=torch.float32)

    return sparse_onehot, unique_values


import pdb


def fast_sparse_dual_iou(onehot1: torch.Tensor, onehot2: torch.Tensor) -> torch.Tensor:
    """
    Returns the (dense) intersection over union between two sparse onehot encoded tensors
    """
    # onehot1 and onehot2 are C1,H*W and C2,H*W

    intersection = torch.sparse.mm(onehot1, onehot2.T).to_dense()
    sparse_sum1 = torch.sparse.sum(onehot1, dim=(1,))[None].to_dense()
    sparse_sum2 = torch.sparse.sum(onehot2, dim=(1,))[None].to_dense()
    union = sparse_sum1.T + sparse_sum2 - intersection

    return (intersection / union)


def iou_test():
    """
    Unit test for the fast dual iou functions
    """
    out = torch.randint(0, 50, (1, 2, 124, 256), dtype=torch.float32)
    onehot1 = torch_onehot(out[0, 0])[0]
    onehot2 = torch_onehot(out[0, 1])[0]
    iou_dense = fast_dual_iou(onehot1, onehot2)

    onehot1 = torch_sparse_onehot(out[0, 0], flatten=True)[0]
    onehot2 = torch_sparse_onehot(out[0, 1], flatten=True)[0]
    iou_sparse = fast_sparse_dual_iou(onehot1, onehot2)

    assert torch.allclose(iou_dense, iou_sparse)



def match_labels(tile_1: torch.Tensor,tile_2: torch.Tensor,threshold: float = 0.5, strict = False):

    """This function takes two labeled tiles, and matches the overlapping labels of tile_2 to the labels of tile_1.
        If strict is set to True, the function will discard non matching objects.
       """
    
    if tile_1.max() == 0 or tile_2.max() == 0:
        if not strict:
            return tile_1, tile_2
        else:
            return torch.zeros_like(tile_1), torch.zeros_like(tile_2)
        
    old_problematic_onehot, old_unique_values = torch_sparse_onehot(tile_1, flatten=True)
    new_problematic_onehot, new_unique_values = torch_sparse_onehot(tile_2, flatten=True)

    iou = fast_sparse_dual_iou(old_problematic_onehot, new_problematic_onehot)

    onehot_remapping = torch.nonzero(iou > threshold).T# + 1

    if old_unique_values.min() == 0:
       old_unique_values = old_unique_values[old_unique_values > 0]
    if new_unique_values.min() == 0:
       new_unique_values = new_unique_values[new_unique_values > 0]


    if onehot_remapping.shape[1] > 0:
        
        onehot_remapping = torch.stack((new_unique_values[onehot_remapping[1]], old_unique_values[onehot_remapping[0]]))

        if not strict:
            mask = torch.isin(tile_2, onehot_remapping[0])
            tile_2[mask] = remap_values(onehot_remapping, tile_2[mask])

            return tile_1, tile_2
        else:
            tile_1 = tile_1 * torch.isin(tile_1, onehot_remapping[1]).int()
            tile_2 = tile_2 * torch.isin(tile_2, onehot_remapping[0]).int()

            tile_2[tile_2>0] = remap_values(onehot_remapping, tile_2[tile_2>0])

            return tile_1, tile_2
        
    else:
        if not strict:
            return tile_1, tile_2
        else:
            return torch.zeros_like(tile_1), torch.zeros_like(tile_2)
        

def connected_components(x: torch.Tensor, num_iterations: int = 32) -> torch.Tensor:
    """
    This function takes a binary image and returns the connected components
    """
    mask = x == 1

    B, _, H, W = x.shape
    out = torch.arange(B * W * H, device=x.device, dtype=x.dtype).reshape((B, 1, H, W))
    out[~mask] = 0

    for _ in range(num_iterations):
        out[mask] = F.max_pool2d(out, kernel_size=3, stride=1, padding=1)[mask]

    return out


def iou_heatmap(x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
    """
    x is H,W
    y is H,W
    This function takes two labeled images and returns the intersection over union heatmap
    """
    if x.max() ==0 or y.max() == 0:
        return torch.zeros_like(x)

    x = torch_fastremap(x)
    y = torch_fastremap(y)

    x_onehot, _ = torch_sparse_onehot(x, flatten=True)
    y_onehot, _ = torch_sparse_onehot(y, flatten=True)

    iou = fast_sparse_dual_iou(x_onehot, y_onehot)
    predicted_iou = iou.sum(1)
    onehot = torch_onehot(x)
    onehot = onehot.float() * predicted_iou[:,None, None]
    map = onehot.max(1)[0]

    return map



def centroids_from_lab(lab: torch.Tensor):
    mesh_grid = torch.stack(torch.meshgrid(torch.arange(lab.shape[-2], device = lab.device), torch.arange(lab.shape[-1],device = lab.device), indexing="ij")).float()

    sparse_onehot, label_ids = torch_sparse_onehot(lab, flatten=True)

    sum_centroids = torch.sparse.mm(sparse_onehot, mesh_grid.flatten(1).T)

    centroids = sum_centroids / torch.sparse.sum(sparse_onehot, dim=(1,)).to_dense().unsqueeze(-1)

    return centroids, label_ids  # N,2  N


def get_patches(lab: torch.Tensor, image: torch.Tensor, patch_size: int = 64, return_lab_ids: bool = False):
    # lab is 1,H,W with N objects
    # image is C,H,W

    # Returns N,C,patch_size,patch_size

    centroids, label_ids = centroids_from_lab(lab)
    N = centroids.shape[0]

    C, h, w = image.shape[-3:]

    window_size = patch_size // 2
    centroids = centroids.clone()  # N,2
    centroids[:, 0] = centroids[:,0].clamp(min=window_size, max=h - window_size)
    centroids[:, 1] = centroids[:,1].clamp(min=window_size, max=w - window_size)
    window_slices = centroids[:, None] + torch.tensor([[-1, -1], [1, 1]]).to(image.device) * window_size
    window_slices = window_slices.long()  # N,2,2

    slice_size = window_size * 2

    # Create grids of indices for slice windows
    grid_x, grid_y = torch.meshgrid(
        torch.arange(slice_size, device=image.device),
        torch.arange(slice_size, device=image.device), indexing="ij")
    mesh = torch.stack((grid_x, grid_y))

    mesh_grid = mesh.expand(N, 2, slice_size, slice_size)  # N,2,2*window_size,2*window_size
    mesh_flat = torch.flatten(mesh_grid, 2).permute(1, 0, -1)  # 2,N,2*window_size*2*window_size
    idx = window_slices[:, 0].permute(1, 0)[:, :, None]
    mesh_flat = mesh_flat + idx
    mesh_flater = torch.flatten(mesh_flat, 1)  # 2,N*2*window_size*2*window_size


    out = image[:, mesh_flater[0], mesh_flater[1]].reshape(C, N, -1)
    out = out.reshape(C, N, patch_size, patch_size)
    out = out.permute(1, 0, 2, 3)


    if return_lab_ids:
        return out, label_ids

    return out,label_ids  # N,C,patch_size,patch_size


def get_masked_patches(lab: torch.Tensor, image: torch.Tensor, patch_size: int = 64):
    # lab is 1,H,W
    # image is C,H,W

    # if lab.max() == 0:
    #     if return_mask:
    #         return None,None
    #     return None

    lab_patches, label_ids = get_patches(lab, lab[0], patch_size)
    mask_patches = lab_patches == label_ids[1:, None, None, None]

    image_patches,_ = get_patches(lab, image, patch_size)

    # canvas = torch.ones_like(image_patches) * (~mask_patches).float()

    # image_patches = image_patches * mask_patches.float() + canvas

   # pdb.set_trace()

    return image_patches,mask_patches  # N,C,patch_size,patch_size

def feature_extractor():
    import torch
    from torchvision.models.resnet import ResNet
    from torchvision.models.resnet import ResNet18_Weights
    import torch.nn as nn
    from typing import Type, Union, List, Optional, Any
    from torchvision.models.resnet import BasicBlock, Bottleneck, WeightsEnum

    class ResNetNoInitialDownsize(ResNet):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.conv1 = nn.Conv2d(3, 64, kernel_size=7, stride=1, padding=3, bias=False)

    def _resnet_custom(
            resnet_constructor,
            block: Type[Union[BasicBlock, Bottleneck]],
            layers: List[int],
            weights: Optional[WeightsEnum],
            progress: bool,
            **kwargs: Any,
    ) -> ResNet:
        # if weights is not None:
        #     _ovewrite_named_param(kwargs, "num_classes", len(weights.meta["categories"]))

        model = resnet_constructor(block, layers, **kwargs)

        if weights is not None:
            model.load_state_dict(weights.get_state_dict(progress=progress))

        return model

    weights = ResNet18_Weights.verify(ResNet18_Weights.IMAGENET1K_V1)
    # weights = None
    model = _resnet_custom(ResNetNoInitialDownsize, BasicBlock, [2, 2, 2, 2], weights, progress=True)

    return model



def eccentricity_batch(mask_tensor):
    """
    Calculate the eccentricity of a batch of binary masks. B,H,W -> returns B
    """
    
    # Get dimensions
    batch_size, m, n = mask_tensor.shape
    
    # Create indices grid
    y_indices, x_indices = torch.meshgrid(torch.arange(m), torch.arange(n), indexing='ij')
    y_indices = y_indices.unsqueeze(0).to(mask_tensor.device).expand(batch_size, m, n)
    x_indices = x_indices.unsqueeze(0).to(mask_tensor.device).expand(batch_size, m, n)
    
    # Find total mass and centroid
    total_mass = mask_tensor.sum(dim=(1, 2))
    centroid_y = (y_indices * mask_tensor).sum(dim=(1, 2)) / total_mass
    centroid_x = (x_indices * mask_tensor).sum(dim=(1, 2)) / total_mass
    
    # Calculate second-order moments
    y_diff = y_indices - centroid_y.view(batch_size, 1, 1)
    x_diff = x_indices - centroid_x.view(batch_size, 1, 1)
    M_yy = torch.sum(y_diff**2 * mask_tensor, dim=(1, 2))
    M_xx = torch.sum(x_diff**2 * mask_tensor, dim=(1, 2))
    M_xy = torch.sum(x_diff * y_diff * mask_tensor, dim=(1, 2))

    # Construct second-order moments tensor
    moments_tensor = torch.stack([torch.stack([M_xx, M_xy]),
                                  torch.stack([M_xy, M_yy])]).permute(2,0,1)
    

    # Compute eigenvalues
    eigenvalues = torch.linalg.eigvals(moments_tensor)

    # Get maximum eigenvalue
    lambda1 = torch.max(eigenvalues.real, dim=1).values
    # Get minimum eigenvalue
    lambda2 = torch.min(eigenvalues.real, dim=1).values
    
    # Calculate eccentricity
    eccentricity = torch.sqrt(1 - (lambda2 / lambda1))
    
    return eccentricity.squeeze(1,2)





import sys
from pathlib import Path
from instanseg.utils.utils import show_images, _move_channel_axis
import fastremap
from tqdm import tqdm
from skimage import io
from pathlib import Path
from scipy import ndimage
import skimage
import numpy as np
import os
import subprocess

import requests
import zipfile


def get_raw_datasets_dir(*others) -> Path:

    if os.environ.get('INSTANSEG_RAW_DATASETS'):
        path = Path(os.environ['INSTANSEG_RAW_DATASETS'])
    else:
        path = Path(os.path.join(os.path.dirname(__file__),"../../Raw_Datasets/"))
        os.environ['INSTANSEG_RAW_DATASETS'] = str(path)
    
    if others:
        path = path.joinpath(*others)

    return path

def get_processed_datasets_dir(*others) -> Path:

    if os.environ.get('INSTANSEG_DATASET_PATH'):
        path = Path(os.environ['INSTANSEG_DATASET_PATH'])
    else:
        path = Path(os.path.join(os.path.dirname(__file__),"../datasets/"))
        os.environ['INSTANSEG_DATASET_PATH'] = str(path)

    if not path.exists():
        path.mkdir(exist_ok=True,parents=True)

    if others:
        path = path.joinpath(*others)

    return path
    

def create_raw_datasets_dir(*others) -> Path:
    path = get_raw_datasets_dir(*others)
    if not path.exists():
        path.mkdir(parents=True, exist_ok=True)
    return path

def create_processed_datasets_dir(*others) -> Path:
    path = get_processed_datasets_dir(*others)
    if not path.exists():
        path.mkdir(parents=True, exist_ok=True)
    return path

    
def load_Cellpose(Segmentation_Dataset: dict, verbose: bool = True) -> dict:

    from instanseg.utils.augmentations import Augmentations
    from instanseg.utils.pytorch_utils import torch_fastremap

    cellpose_dir = create_raw_datasets_dir("Cell_Segmentation", "Cellpose")

    # Process Train Data
    train_file_path = Path(cellpose_dir) / "train"
    train_folders = sorted(list(train_file_path.iterdir()))
    train_items = []

    for file in tqdm(train_folders):
        if "masks" not in file.name:
            item = {}
            image = io.imread(file)
            labels = io.imread(str(file).replace("_img", "_masks"))
            _, area = np.unique(labels[labels > 0], return_counts=True)
            augmenter = Augmentations(shape=(2, 2))
            input_tensor, labels = augmenter.to_tensor(image, labels, normalize=False)
            tensor, labels = augmenter.torch_rescale(input_tensor, labels, current_pixel_size=0.5 / ((np.median(area) ** 0.5) / (300 ** 0.5)), requested_pixel_size=0.5, crop=False)

            item['cell_masks'] = torch_fastremap(labels).squeeze()
            item['image'] = fastremap.refit(np.array(tensor.byte()))
            item["parent_dataset"] = "cellpose"
            item['licence'] = "Non-Commercial"
            item['image_modality'] = "Fluorescence"
            item['file_name'] = file.name
            item['original_size'] = image.shape

            train_items.append(item)

    np.random.seed(42)
    np.random.shuffle(train_items)
    Segmentation_Dataset['Train'] += train_items[:int(len(train_items) * 0.8)]
    Segmentation_Dataset['Validation'] += train_items[int(len(train_items) * 0.8):]
    Segmentation_Dataset['Test'] += train_items[int(len(train_items) * 0.9):]

    # Process Test Data
    test_file_path = Path(cellpose_dir) / "test"
    test_folders = sorted(list(test_file_path.iterdir()))
    test_items = []

    for file in tqdm(test_folders):
        if "masks" not in file.name:
            item = {}
            image = io.imread(file)
            labels = io.imread(str(file).replace("_img", "_masks"))
            _, area = np.unique(labels[labels > 0], return_counts=True)
            augmenter = Augmentations(shape=(2, 2))
            input_tensor, labels = augmenter.to_tensor(image, labels, normalize=False)
            tensor, labels = augmenter.torch_rescale(input_tensor, labels, current_pixel_size=0.5 / ((np.median(area) ** 0.5) / (300 ** 0.5)), requested_pixel_size=0.5, crop=False)
            item['cell_masks'] = torch_fastremap(labels).squeeze()
            item['image'] = fastremap.refit(np.array(tensor.byte()))
            item["parent_dataset"] = "cellpose"
            item['licence'] = "Non-Commercial"
            item['image_modality'] = "Fluorescence"
            item['file_name'] = file.name
            item['original_size'] = image.shape

            test_items.append(item)


    Segmentation_Dataset['Test'] += test_items

    return Segmentation_Dataset

def load_TNBC_2018(Segmentation_Dataset: dict, verbose: bool = True) -> dict:

    tnbc_dir = create_raw_datasets_dir("Nucleus_Segmentation", "TNBC_NucleiSegmentation")
    zip_file_path = tnbc_dir / "TNBC_NucleiSegmentation.zip"
    download_url = "https://zenodo.org/record/3552674/files/TNBC_and_Brain_dataset.zip?download=1"

    if not zip_file_path.exists():
        # Download the dataset using requests
        if verbose:
            print(f"Downloading dataset from {download_url} to {zip_file_path}...")
        response = requests.get(download_url, stream=True)
        response.raise_for_status()
        with open(zip_file_path, 'wb') as f:
            for chunk in response.iter_content(chunk_size=8192):
                f.write(chunk)
        if verbose:
            print(f"Download completed.")

        # Unzip the dataset
        if verbose:
            print(f"Unzipping dataset to {tnbc_dir}...")
        with zipfile.ZipFile(zip_file_path, 'r') as zip_ref:
            zip_ref.extractall(tnbc_dir)
        if verbose:
            print("Unzipping completed.")

    file_path = tnbc_dir / "TNBC_and_Brain_dataset"
    folders = sorted(list(file_path.iterdir()))

    items=[]

    for i,folder in enumerate(tqdm(folders)):
        if "Slide_" in str(folder):
            for file in sorted(os.listdir(folder)):
                file = (folder / file)
                if ".DS_Store" not in str(file):
                    item={}
                    image = io.imread(str(file))
                    if _move_channel_axis(image).shape[0] == 4:  #Remove alpha channel
                        image = _move_channel_axis(image)[:3]

                    masks = io.imread(str(file).replace("Slide_","GT_"))
                    lab, n_labels = ndimage.label(masks > 0)
                    labels, remapping = fastremap.renumber(skimage.morphology.label(lab), in_place=True)

                    item['nucleus_masks']=fastremap.refit(labels)
                    item['image']=image
                    item["parent_dataset"]="TNBC_2018"
                    item['licence']="CC BY 4.0"
                    item['pixel_size']=0.25    #In microns per pixel
                    item['image_modality']="Brightfield"
                    item['stain']="H&E"
                    items.append(item)
                    

    np.random.seed(42) 
    np.random.shuffle(items)
    Segmentation_Dataset['Train']+=items[:int(len(items)*0.8)]
    Segmentation_Dataset['Validation']+=items[int(len(items)*0.8):int(len(items)*0.9)]
    Segmentation_Dataset['Test']+=items[int(len(items)*0.9):]

    return Segmentation_Dataset

def load_LyNSeC(Segmentation_Dataset: dict, verbose: bool = True) -> dict:
    lynsec_dir = create_raw_datasets_dir("Nucleus_Segmentation", "LyNSeC")
    zip_file_path = lynsec_dir / "LyNSeC.zip"
    download_url = "https://zenodo.org/record/8065174/files/lynsec.zip?download=1"

    if not zip_file_path.exists():
        # Download the dataset using requests
        if verbose:
            print(f"Downloading dataset from {download_url} to {zip_file_path}...")
        response = requests.get(download_url, stream=True)
        response.raise_for_status()
        with open(zip_file_path, 'wb') as f:
            for chunk in response.iter_content(chunk_size=8192):
                f.write(chunk)
        if verbose:
            print(f"Download completed.")

        # Unzip the dataset
        if verbose:
            print(f"Unzipping dataset to {lynsec_dir / 'LyNSeC'}...")
        with zipfile.ZipFile(zip_file_path, 'r') as zip_ref:
            zip_ref.extractall(lynsec_dir / 'LyNSeC')
        if verbose:
            print("Unzipping completed.")

    file_path = Path(zip_file_path).parent / "LyNSeC"
    folders = sorted(list(file_path.iterdir()))

    items = []

    for i, folder in enumerate(tqdm(folders)):
        if folder.stem == "lynsec 2":  # Skip incorrectly annotated folder
            continue


        for j, file in enumerate(sorted(os.listdir(folder))):
            file_ = folder / file
            if ".DS_Store" not in str(file_):
                item = {}
                data = np.load(str(file_))
                image = data[:, :, :3].astype(np.uint8)
                masks = data[:, :, 3].copy()
                masks, remapping = fastremap.renumber(masks, in_place=True)
                item['nucleus_masks'] = fastremap.refit(masks)
                item['image'] = image
                item["parent_dataset"] = "LyNSeC"
                item['licence'] = "CC BY 4.0"
                item['pixel_size'] = 0.25  # In microns per pixel
                item['image_modality'] = "Brightfield"
                item['stain'] = "H&E" if i > 1 else "IHC"
                items.append(item)

    np.random.seed(42) 
    np.random.shuffle(items)
    Segmentation_Dataset['Train'] += items[:int(len(items) * 0.8)]
    Segmentation_Dataset['Validation'] += items[int(len(items) * 0.8):int(len(items) * 0.9)]
    Segmentation_Dataset['Test'] += items[int(len(items) * 0.9):]

    return Segmentation_Dataset

def load_NuInsSeg(Segmentation_Dataset: dict, verbose: bool = True) -> dict:
    nuinsseg_dir = create_raw_datasets_dir("Nucleus_Segmentation", "NuInsSeg")
    zip_file_path = nuinsseg_dir / "NuInsSeg.zip"
    download_url = "https://zenodo.org/record/10518968/files/NuInsSeg.zip?download=1"

    if not zip_file_path.exists():
        # Download the dataset using requests
        if verbose:
            print(f"Downloading dataset from {download_url} to {zip_file_path}...")
        response = requests.get(download_url, stream=True)
        response.raise_for_status()
        with open(zip_file_path, 'wb') as f:
            for chunk in response.iter_content(chunk_size=8192):
                f.write(chunk)
        if verbose:
            print(f"Download completed.")

        # Unzip the dataset
        if verbose:
            print(f"Unzipping dataset to {nuinsseg_dir / 'NuInsSeg'}...")
        with zipfile.ZipFile(zip_file_path, 'r') as zip_ref:
            zip_ref.extractall(nuinsseg_dir / 'NuInsSeg')
        if verbose:
            print("Unzipping completed.")


    file_path = nuinsseg_dir / "NuInsSeg"
    folders = sorted(list(Path(file_path).iterdir()))

    items=[]

    for i,folder in enumerate(tqdm(folders)):

        img_folder = folder / "tissue images"

        for j,file in enumerate(sorted(os.listdir(img_folder))):
            file = (img_folder / file)

            if ".DS_Store" not in str(file):
                item={}
                data = io.imread(str(file))
                image = data[:,:,:3].astype(np.uint8)

                masks = io.imread(str(file).replace("tissue images","label masks modify").replace(".png",".tif"))

                masks, remapping = fastremap.renumber(masks ,in_place=True)
                item['nucleus_masks']=fastremap.refit(masks)
                item['image']=image
                item["parent_dataset"]="NuInsSeg"
                item['licence']="CC BY 4.0"
                item['pixel_size']=0.25    #In microns per pixel
                item['image_modality']="Brightfield"
                item['stain']="H&E"
                item["tissue_type"]= folder.stem
                items.append(item)

    np.random.seed(42) 
    np.random.shuffle(items)
    Segmentation_Dataset['Train']+=items[:int(len(items)*0.8)]
    Segmentation_Dataset['Validation']+=items[int(len(items)*0.8):int(len(items)*0.9)]
    Segmentation_Dataset['Test']+=items[int(len(items)*0.9):]

    return Segmentation_Dataset
            

def load_IHC_TMA(Segmentation_Dataset: dict, verbose: bool = True) -> dict:
    ihc_tma_dir = create_raw_datasets_dir("Nucleus_Segmentation", "IHC_TMA_dataset")
    zip_file_path = ihc_tma_dir / "IHC_TMA_dataset.zip"
    download_url = "https://zenodo.org/record/7647846/files/IHC_TMA_dataset.zip?download=1"

    if not zip_file_path.exists():
        # Download the dataset using requests
        if verbose:
            print(f"Downloading dataset from {download_url} to {zip_file_path}...")
        response = requests.get(download_url, stream=True)
        response.raise_for_status()
        with open(zip_file_path, 'wb') as f:
            for chunk in response.iter_content(chunk_size=8192):
                f.write(chunk)
        if verbose:
            print(f"Download completed.")

        # Unzip the dataset
        if verbose:
            print(f"Unzipping dataset to {ihc_tma_dir}...")
        with zipfile.ZipFile(zip_file_path, 'r') as zip_ref:
            zip_ref.extractall(ihc_tma_dir)
        if verbose:
            print("Unzipping completed.")


    file_path = ihc_tma_dir / "IHC_TMA_dataset" / "images"
    files = sorted(list(file_path.iterdir()))

    
    def separate_touching_objects(lab):
        from scipy.ndimage import distance_transform_edt
        from skimage.segmentation import watershed
        from skimage.measure import label

        mask = lab > 0
        distance_map = distance_transform_edt(mask) 
        markers = distance_map > 2
        markers = label(markers)
        labels = watershed(-distance_map, markers, mask=mask)
        labels = labels + lab
        return labels


    items = []

    for i, file in enumerate(tqdm(files)):
        item = {}

        image = io.imread(str(file))
        masks = np.load(str(file).replace("images", "masks").replace(".png", ".npy"))

        n_masks = np.max(masks[0:2], axis=0)

        masks = separate_touching_objects(n_masks)
        masks, remapping = fastremap.renumber(masks, in_place=True)
        masks = fastremap.refit(masks)

        item['nucleus_masks'] = fastremap.refit(masks)
        item['image'] = image
        item["parent_dataset"] = "IHC_TMA"
        item['licence'] = "CC BY 4.0"
        item['pixel_size'] = 0.25  # In microns per pixel
        item['stain'] = "IHC"
        item['image_modality'] = "Brightfield"

        items.append(item)

    np.random.seed(42)
    np.random.shuffle(items)
    Segmentation_Dataset['Train'] += items[:int(len(items) * 0.8)]
    Segmentation_Dataset['Validation'] += items[int(len(items) * 0.8):int(len(items) * 0.9)]
    Segmentation_Dataset['Test'] += items[int(len(items) * 0.9):]

    return Segmentation_Dataset

def load_CoNSeP(Segmentation_Dataset: dict, verbose: bool = True) -> dict:
    consep_dir = create_raw_datasets_dir("Nucleus_Segmentation", "CoNSeP")
    zip_file_path = consep_dir / "CoNSeP.zip"
    download_url = "https://warwick.ac.uk/fac/cross_fac/tia/data/hovernet/consep_dataset.zip"

    if not zip_file_path.exists():
        # Download the dataset using requests
        if verbose:
            print(f"Downloading dataset from {download_url} to {zip_file_path}...")
        response = requests.get(download_url, stream=True)
        response.raise_for_status()
        with open(zip_file_path, 'wb') as f:
            for chunk in response.iter_content(chunk_size=8192):
                f.write(chunk)
        if verbose:
            print(f"Download completed.")

        # Unzip the dataset
        if verbose:
            print(f"Unzipping dataset to {consep_dir / 'CoNSeP'}...")
        with zipfile.ZipFile(zip_file_path, 'r') as zip_ref:
            zip_ref.extractall(consep_dir / 'CoNSeP')
        if verbose:
            print("Unzipping completed.")

    # Process Training Data
    training_path = consep_dir / "CoNSeP" / "Training"
    training_files = sorted(list(training_path.iterdir()))

    items = []

    for i, file in enumerate(tqdm(training_files)):
        if ".DS_Store" not in str(file) and "_masks" not in str(file):
            item = {}
            img = io.imread(file)

            masks = io.imread(str(file.parent / (file.stem + "_masks.tif")))
            masks, remapping = fastremap.renumber(masks, in_place=True)
            masks = fastremap.refit(masks)

            item['nucleus_masks'] = masks
            item['image'] = img
            item["parent_dataset"] = "CoNSeP"
            item['image_modality'] = "Brightfield"
            item['stain'] = "H&E"
            item['licence'] = "Apache 2.0"
            item['pixel_size'] = 0.275
            items.append(item)

    np.random.seed(42)
    np.random.shuffle(items)
    Segmentation_Dataset['Train'] += items[:int(len(items) * 0.8)]
    Segmentation_Dataset['Validation'] += items[int(len(items) * 0.8):]

    # Process Testing Data
    testing_path = consep_dir / "CoNSeP" / "Testing"
    testing_files = sorted(list(testing_path.iterdir()))

    items = []

    for i, file in enumerate(tqdm(testing_files)):
        if ".DS_Store" not in str(file) and "_masks" not in str(file):
            item = {}
            img = io.imread(file)
            masks = io.imread(str(file.parent / (file.stem + "_masks.tif")))
            masks, remapping = fastremap.renumber(masks, in_place=True)
            masks = fastremap.refit(masks)
            item['nucleus_masks'] = masks
            item['image'] = img
            item["parent_dataset"] = "CoNSeP"
            item['image_modality'] = "Brightfield"
            item['stain'] = "H&E"
            item['licence'] = "Apache 2.0"
            item['pixel_size'] = 0.275
            items.append(item)

    Segmentation_Dataset['Test'] += items

    return Segmentation_Dataset

def load_MoNuSeg(Segmentation_Dataset: dict, verbose: bool = True) -> dict:
    monuseg_dir = create_raw_datasets_dir("Nucleus_Segmentation", "MoNuSeg")
    zip_file_path = monuseg_dir / "MoNuSeg.zip"
    download_url = "https://github.com/juglab/EmbedSeg/releases/download/v0.1.0/monuseg-2018.zip"

    if not zip_file_path.exists():
        # Download the dataset using requests
        if verbose:
            print(f"Downloading dataset from {download_url} to {zip_file_path}...")
        response = requests.get(download_url, stream=True)
        response.raise_for_status()
        with open(zip_file_path, 'wb') as f:
            for chunk in response.iter_content(chunk_size=8192):
                f.write(chunk)
        if verbose:
            print(f"Download completed.")

        # Unzip the dataset
        if verbose:
            print(f"Unzipping dataset to {monuseg_dir / 'MoNuSeg'}...")
        with zipfile.ZipFile(zip_file_path, 'r') as zip_ref:
            zip_ref.extractall(monuseg_dir / 'MoNuSeg')
        if verbose:
            print("Unzipping completed.")

    file_path = Path(f"{monuseg_dir}/MoNuSeg/monuseg-2018/download")
    
    # Process Test Data
    items = []
    set_type = "test"
    test_files = sorted(list((file_path / f"{set_type}/images").iterdir()))

    for file in tqdm(test_files):
        if ".DS_Store" not in str(file):
            item = {}
            data = io.imread(str(file))
            image = data[:, :, :3].astype(np.uint8)
            masks = io.imread(str(file).replace("images", "masks"))
            masks, remapping = fastremap.renumber(masks, in_place=True)
            item['nucleus_masks'] = fastremap.refit(masks)
            item['image'] = image
            item["parent_dataset"] = "MoNuSeg"
            item['licence'] = "CC BY NC 4.0"
            item['pixel_size'] = 0.25  # In microns per pixel
            item['image_modality'] = "Brightfield"
            item['stain'] = "H&E"
            items.append(item)

    Segmentation_Dataset['Test'] += items

    # Process Train Data
    items = []
    set_type = "train"
    train_files = sorted(list((file_path / f"{set_type}/images").iterdir()))

    for file in tqdm(train_files):
        if ".DS_Store" not in str(file):
            item = {}
            data = io.imread(str(file))
            image = data[:, :, :3].astype(np.uint8)
            masks = io.imread(str(file).replace("images", "masks"))
            masks, remapping = fastremap.renumber(masks, in_place=True)
            item['nucleus_masks'] = fastremap.refit(masks)
            item['image'] = image
            item["parent_dataset"] = "MoNuSeg"
            item['licence'] = "CC BY NC 4.0"
            item['pixel_size'] = 0.25  # In microns per pixel
            item['image_modality'] = "Brightfield"
            item['stain'] = "H&E"
            items.append(item)

    np.random.seed(42)
    np.random.shuffle(items)
    Segmentation_Dataset['Train'] += items[:int(len(items) * 0.8)]
    Segmentation_Dataset['Validation'] += items[int(len(items) * 0.8):]

    return Segmentation_Dataset



def load_CIL(Segmentation_Dataset: dict, verbose: bool = True) -> dict:
    CIL_dir = create_raw_datasets_dir("Cell_Segmentation", "CIL")
    image_zip_path = CIL_dir / "CIL_images.zip"
    label_zip_path = CIL_dir / "CIL_labels.zip"

    import numpy as np
    import torch
    from skimage import io
    from tqdm import tqdm
    from instanseg.utils.augmentations import Augmentations
    from instanseg.utils.pytorch_utils import torch_fastremap

    if not image_zip_path.exists():
        # Define the URLs and paths
        image_url = "https://cildata.crbs.ucsd.edu/ccdb//telescience/home/CCDB_DATA_USER.portal/P2043/Experiment_6835/Subject_6837/Tissue_6840/Microscopy_6843/MP6843_img_full.zip"
        label_url = "https://cildata.crbs.ucsd.edu/ccdb//telescience/home/CCDB_DATA_USER.portal/P2043/Experiment_6835/Subject_6837/Tissue_6840/Microscopy_6843/MP6843_seg.zip"
        image_extract_path = CIL_dir / "CIL_images"
        label_extract_path = CIL_dir / "CIL_labels"

        # Download the image dataset using requests
        if verbose:
            print(f"Downloading image dataset from {image_url} to {image_zip_path}...")
        response = requests.get(image_url, stream=True)
        response.raise_for_status()
        with open(image_zip_path, 'wb') as f:
            for chunk in response.iter_content(chunk_size=8192):
                f.write(chunk)
        if verbose:
            print(f"Image dataset download completed.")

        # Download the label dataset using requests
        if verbose:
            print(f"Downloading label dataset from {label_url} to {label_zip_path}...")
        response = requests.get(label_url, stream=True)
        response.raise_for_status()
        with open(label_zip_path, 'wb') as f:
            for chunk in response.iter_content(chunk_size=8192):
                f.write(chunk)
        if verbose:
            print(f"Label dataset download completed.")

        # Unzip the image dataset
        if verbose:
            print(f"Unzipping image dataset to {image_extract_path}...")
        with zipfile.ZipFile(image_zip_path, 'r') as zip_ref:
            zip_ref.extractall(image_extract_path)
        if verbose:
            print("Image dataset unzipping completed.")

        # Unzip the label dataset
        if verbose:
            print(f"Unzipping label dataset to {label_extract_path}...")
        with zipfile.ZipFile(label_zip_path, 'r') as zip_ref:
            zip_ref.extractall(label_extract_path)
        if verbose:
            print("Label dataset unzipping completed.")
    train_file_path = Path(CIL_dir) / "CIL_images"
    train_folders = sorted(list(train_file_path.iterdir()))
    items = []

    for file in tqdm(train_folders):
        if "w2" not in file.name:
            item = {}
            image = io.imread(file)
            image2 = io.imread(str(file).replace("w1","w2"))
            img = np.stack([image,image2],axis = 0)

            labels = io.imread(str(file).replace("CIL_images", "CIL_labels").replace("w1.TIF", "_GT_01.tif"))[:,:,0]
            from skimage.measure import label
            labels = label(labels)
            item['cell_masks'] =  labels
            item['image'] = img[:,::2,::2]
            item["parent_dataset"] = "CIL"
            item['licence'] = "CC BY 3.0"
            item['pixel_size'] = 0.3118 * 2
            item['image_modality'] = "Fluorescence"
            item['file_name'] = file.name

            items.append(item)

    np.random.seed(42)
    np.random.shuffle(items)
    Segmentation_Dataset['Train'] += items[:int(len(items) * 0.8)]
    Segmentation_Dataset['Validation'] += items[int(len(items) * 0.8):int(len(items) * 0.9)]
    Segmentation_Dataset['Test'] += items[int(len(items) * 0.9):]

    return Segmentation_Dataset



def download_and_extract(url, dest_path, extract_to, verbose=True):
    if verbose:
        print(f"Downloading dataset from {url} to {dest_path}...")
    response = requests.get(url, stream=True)
    response.raise_for_status()
    with open(dest_path, 'wb') as f:
        for chunk in response.iter_content(chunk_size=8192):
            f.write(chunk)
    if verbose:
        print("Download completed.")

    if verbose:
        print(f"Unzipping dataset to {extract_to}...")
    with zipfile.ZipFile(dest_path, 'r') as zip_ref:
        zip_ref.extractall(extract_to)
    if verbose:
        print("Unzipping completed.")



def load_pannuke(Segmentation_Dataset: dict, verbose: bool = True, no_zip = False) -> dict:
    from pathlib import Path
    import numpy as np
    import fastremap
    import os

    pannuke_dir = create_raw_datasets_dir("Nucleus_Segmentation", "PanNuke")

    if not (pannuke_dir / "fold_1.zip").exists():
        for fold in ["1", "2", "3"]:
            zip_file_path = pannuke_dir / f"fold_{fold}.zip"
            download_url = f"https://warwick.ac.uk/fac/cross_fac/tia/data/pannuke/fold_{fold}.zip"

            # Download and unzip the dataset
            download_and_extract(download_url, zip_file_path, pannuke_dir, verbose)


    processed_pannuke_dir = create_processed_datasets_dir("pannuke_data")


    def get_data(dataset):

        if dataset == "train":
            fold = "1"
        elif dataset == "val":
            fold = "2"
        elif dataset == "test":
            fold = "3"

        import gc
        from tqdm import tqdm
        import tifffile

        out_path = create_processed_datasets_dir("pannuke_data",dataset)

        import fastremap

        items =[]
        masks = np.load(Path(pannuke_dir)/ "Fold {}/masks/fold{}/masks.npy".format(fold,fold), allow_pickle = True).astype(np.int16)
        images = np.load(Path(pannuke_dir)/ "Fold {}/images/fold{}/images.npy".format(fold,fold), allow_pickle = True).astype(np.uint8)
        types = np.load(Path(pannuke_dir)/ "Fold {}/images/fold{}/types.npy".format(fold,fold), allow_pickle = True)

        for i in tqdm(range(len(images))):
            item={}

            assert (np.unique(masks[i,:,:,:-1]) == np.unique(masks[i,:,:,:-1].max(-1))).all()
            label = masks[i,:,:,:-1].max(-1)
            label,_ = fastremap.renumber(label, in_place=True)
            classes = (np.argmax(masks[i,:,:,:-1],axis = -1) + 1) * (label > 0).astype(np.int8)

            image = images[i]

            tifffile.imwrite(out_path / ("image_"+str(i)+".tif"),image)
            tifffile.imwrite(out_path / ("nucleus_masks_"+str(i)+".tif"),fastremap.refit(fastremap.renumber(label)[0]))
            tifffile.imwrite(out_path / ("class_masks_"+str(i)+".tif"),classes)

            relative_path_img = os.path.relpath(str(out_path / ("image_"+str(i)+".tif")),os.environ['INSTANSEG_DATASET_PATH'])
            relative_path_nucleus = os.path.relpath(str(out_path / ("nucleus_masks_"+str(i)+".tif")),os.environ['INSTANSEG_DATASET_PATH'])
            relative_path_class = os.path.relpath(str(out_path / ("class_masks_"+str(i)+".tif")),os.environ['INSTANSEG_DATASET_PATH'])

            item['image']=relative_path_img
            item['nucleus_masks']= relative_path_nucleus
            item['class_masks']= relative_path_class

            item["parent_dataset"]="PanNuke"
            item['licence']="Attribution-NonCommercial-ShareAlike 4.0 International"
            item['image_modality']="Brightfield"
            item['pixel_size']= 0.25   #In microns per pixel
            item['tissue_type']= types[i]
            
            items.append(item)

        return items
    
    Segmentation_Dataset['Train']+=get_data("train")
    Segmentation_Dataset['Validation']+=get_data("val")
    Segmentation_Dataset['Test']+=get_data("test")

    if no_zip:
        return Segmentation_Dataset

    import shutil
    print("Zipping...")#Zip the processed data
    shutil.make_archive(processed_pannuke_dir, 'zip', processed_pannuke_dir)

    return Segmentation_Dataset


def load_tissuenet(Segmentation_Dataset: dict, verbose: bool = True, no_zip = False) -> dict:
    from pathlib import Path
    import numpy as np
    import tifffile
    import gc
    import fastremap
    from tqdm import tqdm
    import os
    import shutil
    import subprocess

    # Create raw and processed dataset directories
    tissuenet_dir = create_raw_datasets_dir("Cell_Segmentation", "TissueNet")
    processed_tissuenet_dir = create_processed_datasets_dir("tissuenet_data")

    def get_data(dataset):
        out_path = processed_tissuenet_dir / dataset
        if not out_path.exists():
            out_path.mkdir(parents=True, exist_ok=True)

        file_path = tissuenet_dir / f"tissuenet_v1.1/tissuenet_v1.1_{dataset}.npz"
        data = np.load(file_path, allow_pickle=True, mmap_mode='r')

        items = []
        imgs = data["X"]
        labels = data["y"]
        metas = data["meta"]

        for i in tqdm(range(len(imgs))):
            item = {}
            image = imgs[i]
            label = labels[i]
            meta = metas[i]

            tifffile.imwrite(out_path / f"image_{i}.tif", fastremap.refit(image))
            tifffile.imwrite(out_path / f"cell_masks_{i}.tif", fastremap.refit(fastremap.renumber(label[:, :, 0])[0]))
            tifffile.imwrite(out_path / f"nucleus_masks_{i}.tif", fastremap.refit(fastremap.renumber(label[:, :, 1])[0]))

            relative_path_img = os.path.relpath(str(out_path / f"image_{i}.tif"), os.environ['INSTANSEG_DATASET_PATH'])
            relative_path_cell = os.path.relpath(str(out_path / f"cell_masks_{i}.tif"), os.environ['INSTANSEG_DATASET_PATH'])
            relative_path_nucleus = os.path.relpath(str(out_path / f"nucleus_masks_{i}.tif"), os.environ['INSTANSEG_DATASET_PATH'])

            item['image'] = relative_path_img
            item['cell_masks'] = relative_path_cell
            item['nucleus_masks'] = relative_path_nucleus

            item["parent_dataset"] = "TissueNet"
            item['licence'] = "Modified Apache, Non-Commercial"
            item['image_modality'] = "Fluorescence"
            item['pixel_size'] = meta[2]  # In microns per pixel
            item['nuclei_channels'] = [0]

            items.append(item)

        data.close()
        del data, imgs, labels, metas
        gc.collect()

        return items

    Segmentation_Dataset['Train'] += get_data("train")
    Segmentation_Dataset['Validation'] += get_data("val")
    Segmentation_Dataset['Test'] += get_data("test")

    if no_zip:
        return Segmentation_Dataset

    print("Zipping...")  # Zip the processed data
    shutil.make_archive(processed_tissuenet_dir, 'zip', processed_tissuenet_dir)

    return Segmentation_Dataset


def load_CPDMI_Vectra(Segmentation_Dataset: dict):

    from pathlib import Path
    from tifffile import TiffFile
    from skimage import io
    import kornia as K
    import torch
    import numpy as np
    import pdb
    import os

    base_path= create_raw_datasets_dir("Cell_Segmentation","CPDMI_2023")

    import pandas as pd
    df=pd.read_excel(base_path/"Annotation Panel Table.xlsx",sheet_name="Vectra")

    VECTRA_dir = create_raw_datasets_dir("Cell_Segmentation","CPDMI_2023", "Vectra")

    subfolders = sorted([path for path in Path(VECTRA_dir).iterdir() if path.is_dir()])
    items=[]
    for subfolder in subfolders:
        subsubfolders = sorted([path for path in Path(subfolder).iterdir() if path.is_dir()])

        for subsubfolder in subsubfolders:
            
            files=sorted([path for path in Path(subsubfolder).iterdir() if path.is_file()])
            cell_mask_path=[path for path in files if "Crop_Cell_Mask_Png" in str(path)]
            nuclei_mask_path=[path for path in files if "Crop_Dapi_Mask_Png" in str(path)]
            img_path=[path for path in files if "Crop_Cell_Tif" in str(path)] + [path for path in files if "Crop_Tif" in str(path)]

            if len(cell_mask_path)==0:
                print("No image found in",subsubfolder)
                continue

            image=io.imread(img_path[0])
            mask=io.imread(cell_mask_path[0])

            if len(nuclei_mask_path)>0:
                nuclei_mask=io.imread(nuclei_mask_path[0])
                connected_nuclei = K.contrib.connected_components(torch.tensor(nuclei_mask[None]).bool().float(), num_iterations=150).int().numpy()
                connected_nuclei = fastremap.renumber(connected_nuclei)[0].squeeze()
                

            if image.shape[-2:]!=mask.shape[-2:]:
                print(image.shape,mask.shape, subfolder,subsubfolder)
                
                continue

            connected = K.contrib.connected_components(torch.tensor(mask[None]).bool().float(), num_iterations=150).int().numpy()
            connected = fastremap.renumber(connected)[0].squeeze()
            connected = fastremap.refit(connected)
            item={}
            item['image']=image
            
            item['cell_masks']=connected

            item["parent_dataset"]="CPDMI_2023"
            item['licence']="CC BY 4.0"
            item['image_modality']="Fluorescence"
            item["platform"]="Vectra"

            if str(subsubfolder).split("/")[-1] == "P13-10002(45819.10401)2250,500":
                continue #this one is corrupted

    
            with TiffFile(img_path[0]) as tif:
                info_string=tif.imagej_metadata['Info']
                info_string=info_string.split("MetaDataPhotometricInterpretation = Monochrome")
                out_channels=info_string[1].split("NewSubfileType = 0")[0].replace("Name #","").split("\n")
                out_channels=[i for i in out_channels if i!=""]
                out_channels=[i.split("=")[1].strip() for i in out_channels]

            with TiffFile(img_path[0]) as tif:
                info_string=tif.imagej_metadata['Info']
                info_list=info_string.split("\n")
                magnification=[i for i in info_list if "XResolution" in i]
                item['resolution']=float(magnification[0].split("=")[1].strip())/1000
                if item['resolution'] > 30:
                    item["pixel_size"] = 0.25
                else:
                    item["pixel_size"] = 0.5
            

            assert len(out_channels)==image.shape[0], print(len(out_channels),image.shape[0])

            item['channel_names']=out_channels
            item['nuclei_channels']=[i for i,val in enumerate(out_channels) if "dapi" in val.lower()]


            if len(nuclei_mask_path)>0:
                connected_nuclei = fastremap.refit(connected_nuclei)

                item['nucleus_masks']=connected_nuclei

            annotation=str(subsubfolder).split("/")[-1].replace("(","[").replace(")","]").replace(".",",") 

            item['filename']=annotation

            if annotation not in df["Annotation:"].unique(): #some annotations are missing
                if annotation == "P07-10004[53024,13505]250,600":
                    item["tissue_type"]="Colon"
                    item["tumor_type"]="Adenocarcinoma"
                elif annotation == "P05-10008[53049,13410]2000,400":
                    item["tissue_type"]="Pancreas"
                    item["tumor_type"]="Pancreatic Ductal Adenocarcinoma"
                elif annotation == "P03-10009[53725,9272]570,10":
                    item["tissue_type"]="Lung"
                    item["tumor_type"]="Adenocarcinoma"
                elif annotation == "P03-10005[46905,11217]800,600":
                    item["tissue_type"]="Lung"
                    item["tumor_type"]="Adenocarcinoma"
                elif annotation == "P03-10005[46905,11217]0,600":
                    item["tissue_type"]="Lung"
                    item["tumor_type"]="Adenocarcinoma"
                elif annotation == "P02-10009[53829,17312]943,604":
                    item["tissue_type"]="Lung"
                    item["tumor_type"]="Adenocarcinoma"

                continue

            item["tissue_type"]=df[df["Annotation:"]==annotation]["Organ:"].values[0]
            item["tumor_type"]=df[df["Annotation:"]==annotation]["Tissue Type:"].values[0]

            items.append(item)

    np.random.seed(42) 
    np.random.shuffle(items)
    Segmentation_Dataset['Train']+=items[:int(len(items)*0.8)]
    Segmentation_Dataset['Validation']+=items[int(len(items)*0.8):]

    return Segmentation_Dataset





def load_CPDMI_Zeiss(Segmentation_Dataset: dict):

    from pathlib import Path
    from tifffile import TiffFile
    from skimage import io
    import kornia as K
    import torch
    import numpy as np

    Zeiss = create_raw_datasets_dir("Cell_Segmentation","CPDMI_2023", "Zeiss")

    subfolders = sorted([path for path in Path(Zeiss).iterdir() if path.is_dir()])
    items=[]
    for subfolder in subfolders:

        files=sorted([path for path in Path(subfolder).iterdir() if path.is_file()])
        cell_mask_path=[path for path in files if "Crop_Cell_Mask_Png" in str(path)]
        img_path=[path for path in files if "Crop_Cell_Tif" in str(path)] + [path for path in files if "Crop_Tif" in str(path)]

        if len(cell_mask_path)==0:
            print("No image found in",subfolder)
            continue

        
        image=io.imread(img_path[0])
        mask=io.imread(cell_mask_path[0])

        if image.shape[-2:]!=mask.shape[-2:]:
            print(image.shape,mask.shape, subfolder,subfolder)
            continue

        connected = K.contrib.connected_components(torch.tensor(mask[None]).bool().float(), num_iterations=150).int().numpy()
        connected = fastremap.renumber(connected)[0].squeeze().astype(np.int16)

        item={}
        item['image']=image
        item['cell_masks']=connected

        item["parent_dataset"]="CPDMI_2023"
        item['licence']="CC BY 4.0"
        item['image_modality']="Fluorescence"
        item["platform"]="Zeiss"
        item["pixel_size"]=0.325

        if "ZP-9999" in subfolder.name:
            item["tissue_type"]="Skin"
            item["tumor_type"]="Basal_Cell"
            item['channel_names']=['DAPI', 'CD8', 'CD4', 'FoxP3', 'PanCK']
            item['nuclei_channels']=[0]
        elif "ZP-10002" in subfolder.name:
            item["tissue_type"]="Skin"
            item["tumor_type"]="CTCL"
            item['channel_names']=['PanCK', 'PD-L1', 'DAPI', 'CD68', 'CD3', 'FoxP3', 'CD8']
            item['nuclei_channels']=[2]
        elif "ZP-10001" in subfolder.name:
            item["tissue_type"]="Skin"
            item["tumor_type"]="CTCL"
            item['channel_names']=['DAPI', 'CD3', 'PD-L1', 'FoxP3', 'PanCK']
            item['nuclei_channels']=[0]
        elif "Spleen" in subfolder.name:
            item["tissue_type"]="Spleen"
            item["tumor_type"]="Melanoma"
            item['channel_names']=['DAPI', 'CD8', 'CD68', 'PD-L1', 'PanCK']
            item['nuclei_channels']=[0]
        elif "PDAC" in subfolder.name:
            item["tissue_type"]="Pancreas"
            item["tumor_type"]="PDAC"
            item['channel_names']=['DAPI', 'CD8', 'CD68', 'PD-L1', 'PanCK']
            item['nuclei_channels']=[0]

        item['filename']=img_path[0].stem

        items.append(item)

    np.random.seed(42) 
    np.random.shuffle(items)
    Segmentation_Dataset['Train']+=items[:int(len(items)*0.8)]
    Segmentation_Dataset['Validation']+=items[int(len(items)*0.8):]

    return Segmentation_Dataset


def load_CPDMI_CODEX(Segmentation_Dataset: dict):

    from pathlib import Path
    from tifffile import TiffFile
    from skimage import io
    import kornia as K
    import torch
    import numpy as np

    CODEX = create_raw_datasets_dir("Cell_Segmentation","CPDMI_2023", "CODEX")

    subfolders = sorted([path for path in Path(CODEX).iterdir() if path.is_dir()])
    items=[]
    for subfolder in subfolders:

        files=sorted([path for path in Path(subfolder).iterdir() if path.is_file()])
        cell_mask_path=[path for path in files if "Crop_Cell_Mask_Png" in str(path)]
        nucleus_mask_path=[path for path in files if "Crop_Dapi_Mask_Png" in str(path)]
        img_path=[path for path in files if "Crop_Cell_Tif" in str(path)] + [path for path in files if "Crop_Tif" in str(path)]

        if len(cell_mask_path)==0:
            print("No image found in",subfolder)
            continue

        image=io.imread(img_path[0])
        mask=io.imread(cell_mask_path[0])

        if len(nucleus_mask_path)>0:
            nuclei_mask=io.imread(nucleus_mask_path[0])
            connected_nuclei = K.contrib.connected_components(torch.tensor(nuclei_mask[None]).bool().float(), num_iterations=150).int().numpy()
            connected_nuclei = fastremap.renumber(connected_nuclei)[0].squeeze().astype(np.int16)

        im=image.transpose(0,3,1,2) 
        image=im.reshape(-1,im.shape[-2],im.shape[-1]) #This was checked by eye and looks correct. The channel ordering was not recorded properly in the metadata.

        if image.shape[-2:]!=mask.shape[-2:]:
            print("Dimension inconsistency",image.shape,mask.shape, subfolder,subfolder)
            continue

        connected = K.contrib.connected_components(torch.tensor(mask[None]).bool().float(), num_iterations=150).int().numpy()
        connected = fastremap.renumber(connected)[0].squeeze().astype(np.int16)
        item={}
        item['image']=image
        item['cell_masks']=connected

        if len(nucleus_mask_path)>0:
            item['nucleus_masks']=connected_nuclei


        item["parent_dataset"]="CPDMI_2023"
        item['licence']="CC BY 4.0"
        item['image_modality']="Fluorescence"
        item["platform"]="CODEX"
        item["pixel_size"]=0.377

        if "LN" in subfolder.name:
            item["tissue_type"]="Lymph_Node"
            item["tumor_type"]="Normal_Tissue"
            item['channel_names']=["DAPI1","NA","NA","NA","DAPI2","CD8","CD3","CD20","DAPI3","Ki67","CD68","PanCK","DAPI4","CD21","CD4","CD31","DAPI5","CD45RO","CD11c","NA","DAPI6","NA","HLA-DR","NA","DAPI7","NA","NA","NA"]
            item['nuclei_channels']=[0, 4, 8, 12, 16, 20, 24]
        elif "Tnsl" in subfolder.name:
            item["tissue_type"]="Tonsil"
            item["tumor_type"]="Normal_Tissue"
            item['channel_names']=['DAPI1', 'NA', 'NA', 'NA', 'DAPI2', 'NA', 'CD4', 'NA', 'API3', 'CD8', 'CD3e', 'CD20', 'DAPI4', 'Ki67', 'HLA-DR', 'DAPI5', 'NA', 'CD68', 'CD31', 'DAPI6', 'CD45RO', 'CD11c', 'NA', 'DAPI7', 'CD21', 'NA', 'NA', 'DAPI8', 'NA', 'NA', 'NA']
            item['nuclei_channels']=[0, 4, 12, 15, 19, 23, 27]

        item['filename']=img_path[0].stem

        items.append(item)

    Segmentation_Dataset['Test']+=items

    return Segmentation_Dataset



def load_BSST265(Segmentation_Dataset: dict, verbose: bool = True) -> dict:
    bsst265_dir = create_raw_datasets_dir("Nucleus_Segmentation", "BSST265")
    zip_file_path = bsst265_dir / "BSST265.zip"
    download_url = "https://www.ebi.ac.uk/biostudies/files/S-BSST265/dataset.zip"
    if not zip_file_path.exists():
        # Download the dataset using requests
        if verbose:
            print(f"Downloading dataset from {download_url} to {zip_file_path}...")
        response = requests.get(download_url, stream=True)
        response.raise_for_status()
        with open(zip_file_path, 'wb') as f:
            for chunk in response.iter_content(chunk_size=8192):
                f.write(chunk)
        if verbose:
            print(f"Download completed.")
        # Unzip the dataset
        if verbose:
            print(f"Unzipping dataset to {bsst265_dir / 'BSST265'}...")
        with zipfile.ZipFile(zip_file_path, 'r') as zip_ref:
            zip_ref.extractall(bsst265_dir / 'BSST265')
        if verbose:
            print("Unzipping completed.")
    import pandas as pd
    metadata = pd.read_csv(bsst265_dir / "BSST265" / "image_description.csv", sep=";")
    
    file_path = Path(f"{bsst265_dir}/BSST265/rawimages")
    files = sorted(list(file_path.iterdir()))
    items = []
    for i, file in enumerate(tqdm(files)):
        metadata_row = metadata[metadata["Image_Name"] == file.stem]
        magnification = metadata_row["Magnification"].values[0]
        if magnification == "20x":
            pixel_size = 0.323
        elif magnification == "40x":
            pixel_size = 0.161
        elif magnification == "63x":
            pixel_size = 0.102
        img_file = str(file)
        mask_file = img_file.replace("rawimages", "groundtruth")
        item = {}
        image = io.imread(img_file)
        masks = io.imread(mask_file)
        item['nucleus_masks'] = fastremap.refit(masks)
        item['image'] = image
        item["parent_dataset"] = "BSST265"
        item['licence'] = "CC0"
        item['pixel_size'] = pixel_size
        item['nuclei_channels'] = [0]  
        items.append(item)
    np.random.seed(42) 
    np.random.shuffle(items)
    Segmentation_Dataset['Train']+=items[:int(len(items)*0.8)]
    Segmentation_Dataset['Validation']+=items[int(len(items)*0.8):int(len(items)*0.9)]
    Segmentation_Dataset['Test']+=items[int(len(items)*0.9):]
    return Segmentation_Dataset



from typing import Tuple, Any

import torch
import pandas as pd
import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt
import sys

import pdb
from instanseg.utils.pytorch_utils import torch_sparse_onehot, torch_fastremap, remap_values, torch_onehot
from instanseg.utils.utils import show_images

def get_intersection_over_union(label: torch.Tensor, return_lab: bool = True) -> torch.Tensor:

    from instanseg.utils.pytorch_utils import fast_sparse_dual_iou, torch_sparse_onehot
    label = torch.stack((torch_fastremap(label[0,0]),torch_fastremap(label[0,1])))[None]
    nucleus_onehot = torch_sparse_onehot(label[0, 0], flatten=True)[0]
    cell_onehot = torch_sparse_onehot(label[0, 1], flatten=True)[0]
    iou = fast_sparse_dual_iou(nucleus_onehot, cell_onehot)
    if return_lab:
        return iou, label
    return iou


def get_intersection_over_nucleus_area(label: torch.Tensor, return_lab: bool = False) -> Tuple[torch.Tensor, torch.Tensor]:
    """
    Returns the intersection over nucleus area in a 2 channel labeled image
     
    label must be a 1,2,H,W tensor where the first channel is nuclei and the second is whole cell
    """
    label = torch.stack((torch_fastremap(label[0,0]),torch_fastremap(label[0,1])))[None]
    nuclei_onehot = torch_sparse_onehot(label[0, 0], flatten=True)[0]
    cell_onehot = torch_sparse_onehot(label[0, 1], flatten=True)[0]
    intersection = torch.sparse.mm(nuclei_onehot, cell_onehot.T).to_dense()
    sparse_sum1 = torch.sparse.sum(nuclei_onehot, dim=(1,))[None].to_dense()
    nuclei_area = sparse_sum1.T

    if return_lab:
        return ((intersection / nuclei_area)), label

    return (intersection / nuclei_area), nuclei_area


def get_intersection_over_cell_area(label: torch.Tensor, return_lab = False) -> Tuple[torch.Tensor, torch.Tensor]:
    """
    Returns the intersection over cell area in a 2 channel labeled image
     
    label must be a 1,2,H,W tensor where the first channel is nuclei and the second is whole cell
    """
    label = torch.stack((torch_fastremap(label[0,0]),torch_fastremap(label[0,1])))[None]
    nuclei_onehot = torch_sparse_onehot(label[0, 0], flatten=True)[0]
    cell_onehot = torch_sparse_onehot(label[0, 1], flatten=True)[0]
    intersection = torch.sparse.mm(cell_onehot, nuclei_onehot.T).to_dense()
    sparse_sum1 = torch.sparse.sum(cell_onehot, dim=(1,))[None].to_dense()
    cell_area = sparse_sum1.T
    if return_lab:
        return ((intersection / cell_area)), label

    return intersection / cell_area, cell_area

def nc_heatmap(label: torch.Tensor) -> torch.Tensor:
    """
    label is a 1,2,H,W tensor where the first channel is nuclei and the second is whole cell
    This function takes two labeled images and returns the nucleus/cell ratio heatmap
    """

    if label[0, 0].max() == 0 or label[0, 1].max() == 0:
        return torch.zeros_like(label[0, 0])


    cells = label[0, 1]

    
    iou, cell_area = get_intersection_over_cell_area(label)
    predicted_iou = iou.sum(1)#[0]
    onehot = torch_onehot(cells)
    onehot = onehot.float() * predicted_iou[:,None, None]
    map = onehot.max(1)[0]

    return map



def get_nonnucleated_cell_ids( lab: torch.Tensor,iou: torch.Tensor = None, threshold: float = 0.5, return_lab: bool = True) -> torch.Tensor:

    if iou is None:
        iou,lab = get_intersection_over_nucleus_area(lab, return_lab=True)
        lab = lab[0,1]
    
    iou = iou > threshold
    lab_ids = torch.unique(lab[lab > 0])
    nonnucleated = (iou.sum(0)) == 0
    if return_lab:
        return lab_ids[nonnucleated], lab * torch.isin(lab, lab_ids[nonnucleated]), lab * torch.isin(lab, lab_ids[
            ~nonnucleated])

    return lab_ids[nonnucleated]


def get_nucleated_cell_ids( lab: torch.Tensor,iou: torch.Tensor = None, threshold: float = 0.5, return_lab: bool = True) -> torch.Tensor:

    if iou is None:
        iou,lab = get_intersection_over_nucleus_area(lab, return_lab=True)
        lab = lab[0,1]
    

    iou = iou > threshold
    lab_ids = torch.unique(lab[lab > 0])
    nucleated = (iou.sum(0)) >= 1

    if return_lab:
        return lab_ids[nucleated], lab * torch.isin(lab, lab_ids[nucleated]), lab * torch.isin(lab, lab_ids[~nucleated])
    return lab_ids[nucleated]


def get_multinucleated_cell_ids( lab: torch.Tensor,iou: torch.Tensor = None, threshold: float = 0.5, return_lab: bool = True):
    #lab is 1,2,H,W
    if iou is None:
        iou,lab = get_intersection_over_nucleus_area(lab, return_lab=True)
        lab = lab[0,1]
    
    iou = iou > threshold
    lab_ids = torch.unique(lab[lab > 0])
    multinucleated = (iou.sum(0)) > 1

    if return_lab:
        return (lab_ids[multinucleated], 
                lab * torch.isin(lab, lab_ids[multinucleated]), 
                lab * torch.isin(lab, lab_ids[~multinucleated]))
    return lab_ids[multinucleated]

def keep_only_largest_nucleus_per_cell(labels: torch.Tensor, return_lab: bool = True)-> Tuple[torch.Tensor, torch.Tensor]:
    """
    labels: tensor of shape 1,2,H,W containing nucleus and cell labels respectively
    return_lab: if True, returns the labels with only the largest nucleus per cell, and only cells that have a nucleus.
    """
    labels = torch_fastremap(labels)
    iou, nuclei_area = get_intersection_over_nucleus_area(labels)
    iou_biggest_area = ((iou > 0.5).float() * nuclei_area) == (((iou > 0.5).float() * nuclei_area).max(0)[0])
    iou_biggest_area = ((iou_biggest_area.float() * iou) > 0.5)
    nuclei_ids = torch.unique(labels[0, 0][labels[0, 0] > 0])
    cell_ids = torch.unique(labels[0, 1][labels[0, 1] > 0])
    largest_nucleus = (iou_biggest_area.sum(1)) == 1
    nucleated_cells = ((iou > 0.5).float().sum(0)) >= 1
    if return_lab:
        return nuclei_ids[largest_nucleus], torch.stack(
            (labels[0, 0] * torch.isin(labels[0, 0], nuclei_ids[largest_nucleus]),
             labels[0, 1] * torch.isin(labels[0, 1], cell_ids[nucleated_cells]))).unsqueeze(0)
    return (nuclei_ids[largest_nucleus],nuclei_ids[largest_nucleus]) #the duplication is to keep torchscript happy



def resolve_cell_and_nucleus_boundaries(lab: torch.Tensor, allow_unnucleated_cells: bool = True) -> torch.Tensor:
    """
    lab: tensor of shape 1,2,H,W containing nucleus and cell labels respectively

    returns: tensor of the same shape as lab

    This function will resolve the boundaries between cells and nuclei. 
    It will first match the labels of the largest nucleus and its cell.
    It will then erase from the cell masks all the nuclei pixels. This resolves nuclei "just" overlapping adjacent cell.
    It will then recover the nuclei pixels that were erased by adding them back to the cell masks.

    allow_unnucleated_cells: If False, this will remove all cells that don't have a nucleus.

    """

    if lab[0,0].max() == 0: # No nuclei
        if lab[0,1].max() == 0 or not allow_unnucleated_cells: # No cells
            return torch.zeros_like(lab)
        else:
            return lab
    elif lab[0,1].max() == 0:
        return torch.stack((lab[0,0],lab[0,0])).unsqueeze(0) #Nuclei but no cells, just duplicate the nuclei.

    lab = torch.stack((torch_fastremap(lab[0, 0]), torch_fastremap(lab[0, 1]))).unsqueeze(0)  # just relabel the nuclei and cells from 1 to N

    original_nuclei_labels = lab[0, 0].clone()
    original_cell_labels = lab[0, 1].clone()
    
    _, lab = keep_only_largest_nucleus_per_cell(lab, return_lab=True) # There will now be as many cells as there are nuclei. But the labels are not yet matched

    lab = torch.stack((torch_fastremap(lab[0, 0]), torch_fastremap(lab[0, 1]))).unsqueeze(0)
     

    if lab[0,0].max() == 0: # No nuclei
        if lab[0,1].max() == 0 or not allow_unnucleated_cells: # No cells
            return torch.zeros_like(lab)
        else:
            return lab
    elif lab[0,1].max() == 0:
        return torch.stack((lab[0,0],lab[0,0])).unsqueeze(0) #Nuclei but no cells, just duplicate the nuclei.
    
    clean_lab = lab 
    
    iou, _ = get_intersection_over_nucleus_area(clean_lab)
    onehot_remapping = (torch.nonzero(iou > 0.5).T + 1).flip(0)
    remapping = torch.cat((torch.zeros(2, 1, device=onehot_remapping.device), onehot_remapping), dim=1)
    clean_lab[0, 1] = remap_values(remapping, clean_lab[0, 1]).int()  # Every matching cell and nucleus now have the same label.


    nuclei_labels = clean_lab[0, 0]
    cell_labels = clean_lab[0, 1]

    original_nuclei_labels[nuclei_labels > 0] = 0
    original_nuclei_labels = torch_fastremap(original_nuclei_labels)
    original_nuclei_labels[original_nuclei_labels > 0]+= nuclei_labels.max()
    nuclei_labels += original_nuclei_labels

    cell_labels[nuclei_labels > 0] = 0
    cell_labels += nuclei_labels

    if allow_unnucleated_cells:

        cell_labels[cell_labels == 0] = (original_cell_labels[cell_labels == 0] + cell_labels.max()) * (original_cell_labels > 0)[cell_labels == 0].float() #this step can create small fragments. This is not a bug - but may have to be cleaned up in the future.
        
    return torch.stack((nuclei_labels, cell_labels)).unsqueeze(0)


def get_mean_object_features(image: torch.Tensor, label: torch.Tensor) -> torch.Tensor:
    # image is C,H,W
    # label is H,W
    # returns a tensor of size N,C for N objects and C channels

    if label.max() == 0:
        return torch.tensor([])
    label = label.squeeze()
    from instanseg.utils.pytorch_utils import torch_sparse_onehot
    sparse_onehot = torch_sparse_onehot(label, flatten=True)[0]
    out = torch.mm(sparse_onehot, image.flatten(1).T)  # object features
    sums = torch.sparse.sum(sparse_onehot, dim=1).to_dense()  # object areas
    out = out / sums[None].T  # mean object features
    return out


def get_features_by_location(input_tensor: torch.Tensor, lab: torch.Tensor, to_numpy: bool = True) -> tuple:
    # input tensor is C,H,W
    # lab is 1,2,H,W where the first channel is nuclei and the second is whole cell

    X_cell = get_mean_object_features(input_tensor, lab[0, 1])
    X_nuclei = get_mean_object_features(input_tensor, lab[0, 0])

    cytoplasm_lab = (lab[0, 0] == 0).float() * lab[0, 1]
    X_nuclei = get_mean_object_features(input_tensor, lab[0, 0])
    X_cytoplasm = get_mean_object_features(input_tensor, cytoplasm_lab)


    if to_numpy:
        X_cell = X_cell.cpu().numpy()
        X_nuclei = X_nuclei.cpu().numpy()
        X_cytoplasm = X_cytoplasm.cpu().numpy()

    return X_cell, X_nuclei, X_cytoplasm


from instanseg.utils.utils import _move_channel_axis
def get_nc_ratio(lab):
    # lab is 1,2,H,W where the first channel is nuclei and the second is whole cell
    
    lab = _move_channel_axis(torch.atleast_3d(lab.squeeze()))[None].long()
    nuclei_mask = lab[0,0]
    cell_mask = lab[0,1]

    nuclei_ids, nuclei_areas = torch.unique(nuclei_mask[nuclei_mask>0].flatten(), return_counts=True)
    cell_ids, cell_areas = torch.unique(cell_mask[cell_mask>0].flatten(), return_counts=True)

    max_labels = cell_mask.max()
    nc_ratio = torch.zeros(max_labels.int(), device=nuclei_mask.device)
    nc_ratio[nuclei_ids - 1] = nuclei_areas.float()
    nc_ratio[cell_ids - 1] /= cell_areas.float()
    nc_ratio = nc_ratio[cell_ids -1 ]
    
    if len(nc_ratio > 1):
        assert nc_ratio.max() <= 1, "Something whent terribly wrong.{}".format(nc_ratio.max())
    

    return nc_ratio




def violin_plot_feature_location(X_nuclei: np.ndarray, X_cytoplasm: np.ndarray, channel_names = None, labels = None, title = None, clamp = None):
    df = pd.DataFrame(columns=['Location', 'Channel', 'Value'])

    for i in range(X_nuclei.shape[1]):
        if channel_names is not None:
            channel_str = channel_names[i]
        else:
            channel_str = str(i)

        if labels is None:
            labels = ['Nuclei', 'Cytoplasm']

        # Assuming you have three arrays: array1, array2, and array3
        # Replace these with your actual data
        array1 = X_nuclei[:, i]
        array2 = X_cytoplasm[:, i]

        if clamp is not None:
            array1 = np.clip(array1, clamp[0], clamp[1])
            array2 = np.clip(array2, clamp[0], clamp[1])

        # Create a DataFrame with the arrays
        df = pd.concat([df, pd.DataFrame({
            'Value': np.concatenate([array1.flatten(), array2.flatten()]),
            ' ': np.repeat(labels, [len(array1), len(array2)]),
            'Channel': np.repeat([channel_str, channel_str], [len(array1), len(array2)])
        })], ignore_index=True)

        # Print the DataFrame

    plt.figure(figsize=(20, 10))
    sns.violinplot(data=df, x="Channel", y="Value", hue=" ", split=True, inner="quart", cut=0)

    if title is not None:
        plt.title(title)
    plt.show()


def show_umap_and_cluster(X_features):
    import scanpy as sc
    # Create a Scanpy AnnData object
    adata = sc.AnnData(X_features)

    # Z-normalise the data
    #sc.pp.scale(adata)

        # Normalizing to median total counts
    sc.pp.normalize_total(adata)
   # sc.pp.scale(adata)
    # Logarithmize the data
    sc.pp.log1p(adata)

    # Perform PCA and UMAP
    # sc.pp.pca(adata)
    sc.pp.neighbors(adata, n_neighbors=5, n_pcs=50)
    sc.tl.umap(adata)

    # Perform leiden clustering
    sc.tl.leiden(adata, resolution=0.1)

    # Plot the UMAP result with leiden clusters
    sc.pl.umap(adata, color='leiden', palette='tab20', legend_loc='on data', title='UMAP with leiden Clustering')

    # Plot the UMAP result with leiden clusters
    plt.show()

    return adata


if __name__ == "__main__":
    from skimage import io
    from instanseg.utils.tiling import sliding_window_inference
    from instanseg.utils.pytorch_utils import torch_sparse_onehot, fast_sparse_dual_iou
    from instanseg.utils.utils import _choose_device
    from instanseg.utils.utils import drag_and_drop_file
    from pathlib import Path

    instanseg = torch.jit.load("../torchscripts/1793450.pt")
    device = _choose_device()
    instanseg.to(device)
    input_data = io.imread("../examples/LuCa1.tif")
   # input_data = io.imread(Path(drag_and_drop_file()))
    from instanseg.utils.augmentations import Augmentations

    Augmenter = Augmentations()
  #  input_tensor, _ = Augmenter.to_tensor(input_data, normalize=True)
    input_tensor,_ =Augmenter.to_tensor(input_data,normalize=False) #this converts the input data to a tensor and does percentile normalization (no clipping)
    input_tensor,_ = Augmenter.normalize(input_tensor)

    print("Running InstanSeg ...")

    lab = instanseg(input_tensor[:,128:256,:128].unsqueeze(0).to(device))

    # lab = sliding_window_inference(input_tensor, instanseg, window_size=(512, 512),
    #                                sw_device=device, device='cpu', output_channels=2)
    
    lab = resolve_cell_and_nucleus_boundaries(lab)

    from instanseg.utils.utils import export_to_torchscript
    from instanseg.utils.model_loader import load_model
    model_str = "1793450"

    device = "cpu"


    model, model_dict = load_model(folder=model_str)
    model.eval()
    model.to(device)

    cells_and_nuclei = model_dict['cells_and_nuclei']
    pixel_size = model_dict['pixel_size']
    n_sigma = model_dict['n_sigma']

    from instanseg.utils.loss.instanseg_loss import InstanSeg_Torchscript
    super_model = InstanSeg_Torchscript(model, cells_and_nuclei=cells_and_nuclei, 
                                        pixel_size = pixel_size, 
                                        n_sigma = n_sigma, 
                                        feature_engineering_function = str(model_dict["feature_engineering"]), 
                                        backbone_dim_in= 3, 
                                        to_centre = bool(model_dict["to_centre"]),
                                        mixed_precision = False).to(device)
    out = super_model(input_tensor[None,][:,:,128:256,:128])



    pdb.set_trace()





    export_to_torchscript(model_name, show_example=True, mixed_predicision=False)






    print("Calculating cellular features ...")
    X_cell, X_nuclei, X_cytoplasm = get_features_by_location(input_tensor, lab)

    print("Plotting violing plots ...")

    violin_plot_feature_location(X_nuclei, X_cytoplasm)

    print("Clustering and umap ...")

    show_umap_and_cluster(X_cell)

    print("Done !")
